/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Bernoulli mixture model classifier.
 */


#include <config.h>

#include <stdlib.h>
#include <stdio.h>
#include <inttypes.h>
#include <assert.h>
#include <string.h>
#include <errno.h>
#include <math.h>

#include <jwsc/base/error.h>
#include <jwsc/base/limits.h>
#include <jwsc/base/string.h>
#include <jwsc/base/file_io.h>
#include <jwsc/prob/pdf.h>
#include <jwsc/vector/vector.h>
#include <jwsc/vector/vector_io.h>
#include <jwsc/vector/vector_math.h>
#include <jwsc/matrix/matrix.h>
#include <jwsc/matrix/matrix_math.h>
#include <jwsc/matrix/matrix_io.h>
#include <jwsc/stat/stat.h>
#include <jwsc/stat/bmm.h>

#include "haplo_groups.h"
#include "bmm.h"


/**
 * @param  model_haplos_out        Haplogroups for the model in the info entry.
 * @param  model_markers_out       Markers for the model in the info entry.
 * @param  model_priors_out        Priors over the haplogroups in the info
 *                                 entry. Only use if @em
 *                                 *model_priors_given_out == 1.
 * @param  model_priors_given_out  Result parameter. Boolean value indicates 
 *                                 whether or not the @em model_priors_out 
 *                                 result parameter was initialized from the
 *                                 info file and should be used.
 * @param  model_fname_out         Name of the model for info file entry read.
 * @param  num_mix_out             Number of mixture components.
 * @param  haplos                  Haplogroups to use for constructing the 
 *                                 subset used for the model in the info entry.
 * @param  markers                 Markers to use for constructing the 
 *                                 subset used for the model in the info entry.
 * @param  fp                      File stream to read the entry from.
 *
 * @return On success, NULL is returned; otherwise NULL is returned but does not
 * free result parameters. 
 */
static Error* read_model_info_entry
(
    Vector_u32**      model_haplos_out,
    Matrix_i32**      model_markers_out,
    Vector_d**        model_priors_out,
    uint8_t*          model_priors_given_out,
    String**          model_fname_out,
    uint32_t*         num_mix_out,
    const Vector_u32* haplos,
    const Matrix_i32* markers,
    FILE*             fp
)
{
    uint32_t num_haplogroups;
    uint32_t num_haplos;
    uint32_t h;
    uint32_t num_aliases;
    uint32_t a;
    uint32_t num_samples;
    uint32_t sample;
    uint32_t num_samples_out;
    uint32_t num_markers;
    uint32_t marker;
    uint32_t haplo;
    uint32_t label;
    uint32_t alias;
    uint32_t model_priors_given;
    uint8_t  found;
    char     haplo_buf[256] = {0};
    char     aliases_buf[1024] = {0};
    char     fname_buf[256] = {0};
    char*    buf_ptr;

    const char* alias_name;
    const char* haplo_name;

    Vector_u32* model_haplos;
    Matrix_i32* model_markers;
    Vector_d*   model_priors;
    String*     model_fname;
    String*     str = NULL;
    Error*      e;

    Vector_u32*  haplo_alias_labels = NULL;
    Vector_u32** haplo_alias        = NULL;

    num_haplogroups = get_num_haplo_groups();

    if (fscanf(fp, "%255s %u %u %u\n", fname_buf, &num_haplos,
                num_mix_out, &model_priors_given) != 4)
    {
        return create_error(ERROR_INV_ARG, __FILE__, "read_model_info_entry",
                __LINE__, "Model information file not formatted properly");
    }
    *model_priors_given_out = (model_priors) ? 1 : 0;

    create_string_from_char(model_fname_out, fname_buf);
    model_fname = *model_fname_out;

    num_samples = markers->num_rows;
    num_markers = markers->num_cols;

    create_vector_u32(model_haplos_out, num_samples);
    model_haplos = *model_haplos_out;

    create_matrix_i32(model_markers_out, num_samples, num_markers);
    model_markers = *model_markers_out;

    if (model_priors_given)
    {
        create_vector_d(model_priors_out, num_haplogroups);
        model_priors = *model_priors_out;
    }


    create_vector_u32(&haplo_alias_labels, num_haplos);
    assert(haplo_alias = malloc(num_haplos*sizeof(Vector_u32)));

    for (h = 0; h < num_haplos; h++)
    {
        if (fscanf(fp, "%255s", haplo_buf) != 1)
        {
            return create_error(ERROR_INV_ARG, __FILE__, 
                    "read_model_info_entry", __LINE__, 
                    "Model information file not formatted properly");
        }

        if ((e = lookup_haplo_group_index_from_label(
                        &(haplo_alias_labels->elts[ h ]),
                        (const char*)haplo_buf)) != NULL)
        {
            return e;
        }

        if (fscanf(fp, "%1023s", aliases_buf) != 1)
        {
            return create_error(ERROR_INV_ARG, __FILE__, 
                    "read_model_info_entry", __LINE__, 
                    "Model information file not formatted properly");
        }

        num_aliases = 1;
        buf_ptr = aliases_buf;
        while (*buf_ptr != '\0')
        {
            if (*buf_ptr == ':')
            {
                *buf_ptr = '\0';
                num_aliases++;
            }
            buf_ptr++;
        }

        haplo_alias[ h ] = NULL;
        create_vector_u32(&(haplo_alias[ h ]), num_aliases);

        buf_ptr = aliases_buf;
        for (a = 0; a < num_aliases; a++)
        {
            assert(sscanf(buf_ptr, "%255s", haplo_buf) == 1);
            if ((e = lookup_haplo_group_index_from_label(
                            &(haplo_alias[ h ]->elts[ a ]),
                            (const char*)haplo_buf)) != NULL)
            {
                return e;
            }

            while (*buf_ptr != '\0')
            {
                buf_ptr++;
            }
            buf_ptr++;
        }

        if (model_priors_given)
        {
            label = haplo_alias_labels->elts[ h ];
            if (fscanf(fp, "%lf", &(model_priors->elts[ label ])) != 1)
            {
                return create_error(ERROR_INV_ARG, __FILE__, 
                        "read_model_info_entry", __LINE__, 
                        "Model information file not formatted properly");
            }
        }
    }


    num_samples_out = 0;

    for (sample = 0; sample < num_samples; sample++)
    {
        haplo = haplos->elts[ sample ];
        found = 0;

        for (h = 0; !found && h < num_haplos; h++)
        {
            label = haplo_alias_labels->elts[ h ];
            num_aliases = haplo_alias[ h ]->num_elts;

            for (a = 0; !found && a < num_aliases; a++)
            {
                alias = haplo_alias[ h ]->elts[ a ];

                assert(lookup_haplo_group_label_from_index(&alias_name, alias) 
                        == NULL);
                assert(lookup_haplo_group_label_from_index(&haplo_name, haplo) 
                        == NULL);

                if (strncmp(alias_name, haplo_name, strlen(alias_name)) == 0)
                {
                    found = 1;
                }
            }
        }

        if (found)
        {
            model_haplos->elts[ num_samples_out ] = label;
            for (marker = 0; marker < num_markers; marker++)
            {
                model_markers->elts[ num_samples_out ][ marker ] =
                    markers->elts[ sample ][ marker ];
            }
            num_samples_out++;
        }
    }


    if (num_samples_out == 0)
    {
        string_printf(&str, 256, "No haplo samples to train model '%s'", 
                model_fname->chars);
        return create_error(ERROR_INV_ARG, __FILE__, "read_model_info_entry", 
                __LINE__, str->chars);
    }


    *model_haplos_out = NULL;
    assert(copy_vector_section_u32(model_haplos_out, model_haplos, 0, 
                num_samples_out) == NULL);
    free_vector_u32(model_haplos);

    *model_markers_out = NULL;
    assert(copy_matrix_block_i32(model_markers_out, model_markers, 0, 0, 
                num_samples_out, num_markers) == NULL);
    free_matrix_i32(model_markers);


    free_vector_u32(haplo_alias_labels);
    for (h = 0; h < num_haplos; h++)
    {
        free_vector_u32(haplo_alias[ h ]);
    }
    free(haplo_alias);

    return NULL;
}


/**
 * @param  model_out     Result parameter. If @em *model_out is NULL, a
 *                       model is allocated; otherwise its space is re-used.
 * @param  haplos        Haplogroups for training, with ith element as 
 *                       corresponding to the ith sample in @em markers.
 * @param  markers       Markers for training, with ith row as a sample 
 *                       corresponding to the ith haplo in @em haplos.
 * @param  haplo_priors  Priors for each of the haplogroups.
 * @param  num_mix       Number of mixture components.
 */
void train_berno_mixture_model
(
    Berno_mixture_model** model_out,
    const Vector_u32*     haplos, 
    const Matrix_i32*     markers,
    const Vector_d*       haplo_priors,
    uint32_t              num_mix
)
{
    uint32_t  h, num_haplogroups;
    uint32_t  s, num_samples;
    uint32_t  m, num_markers;
    uint32_t  n, N;

    Matrix_d**   mu;
    Vector_d**   pi;
    Vector_d*    priors;
    Matrix_u32*  markers_u32 = NULL;

    assert(markers->num_rows == haplos->num_elts);

    num_markers = markers->num_cols;
    num_samples = markers->num_rows;
    num_haplogroups = get_num_haplo_groups();

    if (*model_out != NULL)
    {
        assert((*model_out)->num_markers == num_markers);
        assert((*model_out)->num_haplogroups == num_haplogroups);
        assert((*model_out)->num_mix == num_mix);
        mu     = (*model_out)->mu;
        pi     = (*model_out)->pi;
        priors = (*model_out)->haplo_priors;
    }
    else
    {
        mu     = NULL;
        pi     = NULL;
        priors = NULL;
    }

    assert(mu = malloc(num_haplogroups*sizeof(Matrix_d*)));
    assert(pi = malloc(num_haplogroups*sizeof(Vector_d*)));

    for (h = 0; h < num_haplogroups; h++)
    {
        mu[ h ] = NULL;
        pi[ h ] = NULL;

        N = 0;
        for (s = 0; s < num_samples; s++)
        {
            if (haplos->elts[ s ] == h)
            {
                N++;
            }
        }

        // TODO we should probably handle the case of N=1 by creating a
        // distribution with the sample as the mean and a very small diagonal
        // covariance matrix.
        if (N < num_mix*2)
            continue;

        create_matrix_u32(&markers_u32, N, num_markers);

        n = 0;
        for (s = 0; s < num_samples; s++)
        {
            if (haplos->elts[ s ] == h)
            {
                for (m = 0; m < num_markers; m++)
                {
                    assert(markers->elts[ s ][ m ] >= 0);
                    markers_u32->elts[ n ][ m ] = 
                        (uint32_t)markers->elts[ s ][ m ];
                }
                n++;
            }
        }

        train_bmm_d(&(mu[ h ]), &(pi[ h ]), markers_u32, num_mix);
    }

    free_matrix_u32(markers_u32);

    if (haplo_priors == NULL)
    {
        create_zero_vector_d(&priors, num_haplogroups);

        for (s = 0; s < num_samples; s++)
        {
            h = haplos->elts[ s ];
            priors->elts[ h ]++;
        }
        normalize_vector_sum_d(&priors, priors);
    }
    else
    {
        normalize_vector_sum_d(&priors, haplo_priors);
    }

    if (*model_out == NULL)
    {
        assert(*model_out = malloc(sizeof(Berno_mixture_model)));
    }
    (*model_out)->num_markers     = num_markers;
    (*model_out)->num_haplogroups = num_haplogroups;
    (*model_out)->num_mix         = num_mix;
    (*model_out)->mu              = mu;
    (*model_out)->pi              = pi;
    (*model_out)->haplo_priors    = priors;
}




/**
 * The models are written to the file names specified in the model information
 * file.
 *
 * @param  haplos         Haplogroups for training, with ith element as 
 *                        corresponding to the ith sample in @em markers.
 * @param  markers        Markers for training, with ith row as a sample 
 *                        corresponding to the ith haplo in @em haplos.
 * @param  info_fname     File containing the model information.
 * @param  model_dirname  Directory to write the models to. If NULL, the current
 *                        working directory is used.
 *
 * @return On success, NULL is returned; otherwise an error is returned.
 */
Error* train_berno_mixture_model_info
(
    const Vector_u32* haplos, 
    const Matrix_i32* markers,
    const char*       info_fname,
    const char*       model_dirname
)
{
    FILE* fp;

    Berno_mixture_model* model;
    uint32_t             num_mix;

    Vector_u32*     model_haplos;
    Matrix_i32*     model_markers;
    Vector_d*       model_priors;
    uint8_t         model_priors_given;
    String*         model_fname;
    String*         str;
    Error*          e;
    char            fname_buf[1024];


    model         = NULL;
    model_haplos  = NULL;
    model_markers = NULL;
    model_priors  = NULL;
    model_fname   = NULL;
    str           = NULL;

    if ((fp = fopen(info_fname, "r")) == NULL)
    {
        return create_error(ERROR_IO, __FILE__, 
                "train_berno_mixture_model_info", __LINE__,
                strerror(errno));
    }

    while (skip_fp_spaces_and_comments(fp))
    {
        model_priors_given = 0;

        if ((e = read_model_info_entry(&model_haplos, &model_markers,
                        &model_priors, &model_priors_given, &model_fname,
                        &num_mix, haplos, markers, fp)) != NULL)
        {
            return e;
        }

        train_berno_mixture_model(&model, model_haplos, model_markers,
                model_priors_given ? model_priors : NULL, num_mix);

        if (model_dirname)
        {
            snprintf(fname_buf, 1023, "%s/%s", model_dirname, 
                    model_fname->chars);
        }
        else
        {
            snprintf(fname_buf, 1023, "%s", model_fname->chars);
        }

        if ((e = write_berno_mixture_model(model, fname_buf)) != NULL)
        {
            string_printf(&str, 256, "%s: %s", model_fname->chars, e->msg);
            e = create_error(ERROR_IO, __FILE__, 
                    "train_berno_mixture_model_info", __LINE__, str->chars);
            return e;
        }
    }

    free_berno_mixture_model(model);
    free_vector_u32(model_haplos);
    free_matrix_i32(model_markers);
    free_vector_d(model_priors);
    free_string(model_fname);

    if (fclose(fp) != 0)
    {
        return create_error(ERROR_IO, __FILE__, 
                "train_berno_mixture_model_info", __LINE__, strerror(errno));
    }

    return NULL;
}




/**
 * @param  model  Model to write.
 * @param  fname  File to write to.
 *
 * @return On success, NULL is returned; otherwise an error is returned.
 */
Error* write_berno_mixture_model(Berno_mixture_model* model, const char* fname)
{
    uint32_t h;
    FILE* fp;

    if ((fp = fopen(fname, "w")) == NULL)
    {
        return create_error(ERROR_IO, __FILE__, "write_berno_mixture_model",
                __LINE__, strerror(errno));
    }

    fprintf(fp, "%u\n", model->num_markers);
    fprintf(fp, "%u\n", model->num_haplogroups);
    fprintf(fp, "%u\n", model->num_mix);

    for (h = 0; h < model->num_haplogroups; h++)
    {
        fprintf(fp, "%u ", (model->mu[ h ] != NULL));
    }
    fprintf(fp, "\n");

    for (h = 0; h < model->num_haplogroups; h++)
    {
        if (model->mu[ h ])
        {
            write_matrix_with_header_fp_d(model->mu[ h ], fp);
            write_vector_with_header_fp_d(model->pi[ h ], fp);
        }
    }

    write_vector_with_header_fp_d(model->haplo_priors, fp);

    if (fclose(fp) != 0)
    {
        return create_error(ERROR_IO, __FILE__, "write_berno_mixture_model", 
                __LINE__, strerror(errno));
    }

    return NULL;
}




/**
 * @param  model_out  Result parameter. If @em *model_out is NULL, a model is 
 *                    allocated; otherwise its space is re-used.
 * @param  fname      File to read the model from.
 *
 * @return On success, NULL is returned; otherwise an error is returned but does
 * not free result parameter.
 */
Error* read_berno_mixture_model
(
    Berno_mixture_model** model_out, 
    const char*           fname
)
{
    uint32_t h;

    FILE*                fp;
    Berno_mixture_model* model;
    Error*               e;

    if ((fp = fopen(fname, "r")) == NULL)
    {
        return create_error(ERROR_IO, __FILE__, "read_berno_mixture_model", 
                __LINE__, strerror(errno));
    }

    if(*model_out == NULL)
    {
        assert((*model_out = malloc(sizeof(Berno_mixture_model))) != NULL);
        model = *model_out;
        model->haplo_priors = NULL;
    }

    if (fscanf(fp, "%u\n%u\n%u\n", &(model->num_markers),
                &(model->num_haplogroups), &(model->num_mix)) != 3)
    {
        return create_error(ERROR_INV_ARG, __FILE__, "read_berno_mixture_model",
                __LINE__, "Improperly formatted model file");
    }

    assert(model->mu = malloc(model->num_haplogroups*sizeof(Matrix_d*)));
    assert(model->pi = malloc(model->num_haplogroups*sizeof(Vector_d*)));

    for (h = 0; h < model->num_haplogroups; h++)
    {
        if (fscanf(fp, "%u", (uint32_t*)(&(model->mu[ h ]))) != 1)
        {
            return create_error(ERROR_INV_ARG, __FILE__, 
                    "read_berno_mixture_model", __LINE__, 
                    "Improperly formatted model file");
        }
    }

    for (h = 0; h < model->num_haplogroups; h++)
    {
        model->pi[ h ] = NULL;

        if (model->mu[ h ])
        {
            model->mu[ h ] = NULL;

            if ((e = read_matrix_with_header_fp_d(&(model->mu[h]), fp)) ||
                (e = read_vector_with_header_fp_d(&(model->pi[h]), fp)))
            {
                return create_error(ERROR_INV_ARG, __FILE__, 
                        "read_berno_mixture_model", __LINE__, 
                        "Improperly formatted model file");
            }
        }
    }

    if ((e = read_vector_with_header_fp_d(&(model->haplo_priors), fp)) != NULL)
    {
        return create_error(ERROR_INV_ARG, __FILE__, "read_berno_mixture_model",
                __LINE__, "Improperly formatted model file");
    }

    if (fclose(fp) != 0)
    {
        return create_error(ERROR_IO, __FILE__, "read_berno_mixture_model", 
                __LINE__, strerror(errno));
    }

    return NULL;
}




/** @param  model  Model to free. */
void free_berno_mixture_model(Berno_mixture_model* model)
{
    uint32_t h;

    if (!model)
        return;

    for (h = 0; h < model->num_haplogroups; h++)
    {
        free_matrix_d(model->mu[ h ]);
        free_vector_d(model->pi[ h ]);
    }

    free(model->mu);
    free(model->pi);

    free_vector_d(model->haplo_priors);

    free(model);
}




/** 
 * @param  haplo_out       Result parameter. 
 * @param  confidence_out  Result parameter. 
 * @param  markers         Marker data to predict.
 * @param  model           Trained model to use for predicting.
 *
 * @return On success, NULL is returned; otherwise an error is returned.
 */
Error* predict_haplo_with_berno_mixture_model
(
    uint32_t*                  haplo_out,
    double*                    confidence_out,
    const Vector_i32*          markers,
    const Berno_mixture_model* model
)
{
    uint32_t m, num_markers;
    uint32_t h, best_h, num_haplogroups;
    uint32_t i;

    double log_prior;
    double log_map;
    double log_ll;
    double log_posterior;
    double posterior_sum;

    Matrix_u32* markers_u32 = NULL;


    static const char* haplo_not_found = "X"; 


    num_haplogroups = get_num_haplo_groups();
    num_markers     = markers->num_elts;

    assert(num_haplogroups == model->num_haplogroups);
    assert(num_markers == model->num_markers);

    log_map = log(0);
    posterior_sum = 0;

    assert(lookup_haplo_group_index_from_label(&best_h, haplo_not_found) 
            == NULL);

    create_matrix_u32(&markers_u32, 1, num_markers);

    for(h = 0; h < num_haplogroups; h++)
    {
        if (!model->mu[ h ])
            continue;

        for(m = 0; m < num_markers; m++)
        {
            if (markers->elts[ m ] < 0)
            {
                // Impute missing value with the mean.
                // TODO computation is not correct here
                markers_u32->elts[0][ m ] = 0;
                for (i = 0; i < model->num_mix; i++)
                {
                    markers_u32->elts[0][ m ] += 
                        model->mu[ h ]->elts[ i ][ m ] *
                        model->pi[ h ]->elts[ i ];
                }
            }
            else
            {
                markers_u32->elts[0][ m ] = (uint32_t)markers->elts[ m ];
            }
        }

        log_ll = bmm_log_likelihood_d(model->mu[ h ], model->pi[ h ], 
                markers_u32);

        if (model->haplo_priors->elts[ h ] < JWSC_MIN_LOG_ARG)
        {
            log_prior = log(JWSC_MIN_LOG_ARG);
        }
        else
        {
            log_prior = log(model->haplo_priors->elts[ h ]);
        }

        log_posterior = log_ll + log_prior;
        posterior_sum += exp(log_posterior);

        if (log_posterior > log_map)
        {
            log_map = log_posterior;
            best_h = h;
        }
    }

    if (posterior_sum < JWSC_MIN_LOG_ARG)
    {
        *haplo_out = best_h;
        *confidence_out = 0;
    }
    else
    {
        log_map -= log(posterior_sum);

        *haplo_out = best_h;
        *confidence_out = exp(log_map);
    }

    free_matrix_u32(markers_u32);

    return NULL;
}




/** 
 * @param  haplos_out      Result parameter. If @em *haplos_out is NULL, it is
 *                         allocated; otherwise its space is re-used.
 * @param  confidence_out  Result parameter. If @em *confidence_out is NULL, it 
 *                         is allocated; otherwise its space is re-used.
 * @param  markers         Marker data to predict. Each row is a sample for
 *                         prediction, corresponding to an element in the result
 *                         parameters.
 * @param  model           Trained model to use for predicting.
 *
 * @return On success, NULL is returned; otherwise an error is returned, but the
 * result parameters are not freed.
 */
Error* predict_haplos_with_berno_mixture_model
(
    Vector_u32**              haplos_out,
    Vector_d**                confidence_out,
    const Matrix_i32*         markers,
    const Berno_mixture_model* model
)
{
    uint32_t s, num_samples;
    uint32_t m, num_markers;

    Vector_u32* haplos;
    Vector_d*   confidence;
    Vector_i32* markers_v = NULL;
    Error*      e;

    num_samples = markers->num_rows;
    num_markers = markers->num_cols;

    assert(num_markers == model->num_markers);

    create_vector_u32(haplos_out, num_samples);
    haplos = *haplos_out;

    create_vector_d(confidence_out, num_samples);
    confidence = *confidence_out;

    create_vector_i32(&markers_v, num_markers);

    for(s = 0; s < num_samples; s++)
    {
        for(m = 0; m < num_markers; m++)
        {
            markers_v->elts[ m ] = markers->elts[ s ][ m ];
        }

        if ((e = predict_haplo_with_berno_mixture_model(&(haplos->elts[ s ]), 
                        &(confidence->elts[ s ]), markers_v, model)))
        {
            return e;
        }
    }

    free_vector_i32(markers_v);

    return NULL;
}



/** @brief Recursively predicts a haplogroup tree. */
static Error* recursively_predict_haplo_in_model_tree
(
    uint32_t*                       haplo_out,
    double*                         confidence_out,
    const Berno_mixture_model_tree* tree,
    const Vector_i32*               markers
)
{
    uint32_t subtree_haplo;
    uint32_t subtree;

    double confidence;

    Error* e;

    if ((e = predict_haplo_with_berno_mixture_model(haplo_out, &confidence, 
                    markers, tree->model)))
    {
        return e;
    }

    *confidence_out *= confidence;

    /* If the best model has a subtree, recursively predict in the subtree. */
    for (subtree = 0; subtree < tree->num_subtrees; subtree++)
    {
        subtree_haplo = tree->subtrees[ subtree ]->parent_haplo;

        if (subtree_haplo == *haplo_out)
        {
            if ((e = recursively_predict_haplo_in_model_tree(haplo_out,
                            confidence_out, tree->subtrees[ subtree ], 
                            markers)))
            {
                return e;
            }
        }
    }

    return NULL;
}


/** 
 * @param  haplos_out      Result parameter. If @em *haplos_out is NULL, it is
 *                         allocated; otherwise its space is re-used.
 * @param  confidence_out  Result parameter. If @em *confidence_out is NULL, it 
 *                         is allocated; otherwise its space is re-used.
 * @param  markers         Marker data to predict. Each row is a sample for
 *                         prediction, corresponding to an element in the result
 *                         parameters.
 * @param  treel           Trained model tree to use for predicting.
 *
 * @return On success, NULL is returned; otherwise an error is returned, but the
 * result parameters are not freed.
 */
Error* predict_haplos_with_berno_mixture_model_tree
(
    Vector_u32**                    haplos_out,
    Vector_d**                      confidence_out,
    const Matrix_i32*               markers,
    const Berno_mixture_model_tree* tree
)
{
    uint32_t s, num_samples;
    uint32_t m, num_markers;

    Vector_u32* haplos;
    Vector_d*   confidence;
    Vector_i32* markers_v = NULL;
    Error*      e;

    num_samples = markers->num_rows;
    num_markers = markers->num_cols;

    assert(num_markers == tree->model->num_markers);

    create_vector_u32(haplos_out, num_samples);
    haplos = *haplos_out;

    create_init_vector_d(confidence_out, num_samples, 1.0);
    confidence = *confidence_out;

    create_vector_i32(&markers_v, num_markers);

    for (s = 0; s < num_samples; s++ )
    {
        for (m = 0; m < num_markers; m++ )
        {
            markers_v->elts[ m ]= markers->elts[ s ][ m ];
        }

        if ((e = recursively_predict_haplo_in_model_tree(&(haplos->elts[ s ]), 
                        &(confidence->elts[ s ]), tree, markers_v)))
        {
            return e;
        }
    }

    free_vector_i32(markers_v);

    return NULL;
}




/** 
 * @brief Recursively reads a subtree of a model tree.
 *
 * @param  subtree_out    Result parameter. If @em *tree_out is NULL, a model 
 *                        tree is allocated; otherwise it is first freed.
 * @param  parent         Parent of the subtree being read.
 * @param  fp             File stream to read the model subtree from.
 * @param  model_dirname  Directory to read the models from. If NULL, the
 *                        current working directory is used.
 *
 * @return On success, NULL is returned; otherwise an error is returned, but
 * the result parameters are not freed.
 */
static Error* read_berno_mixture_model_subtree
(
    Berno_mixture_model_tree**      subtree_out,
    const Berno_mixture_model_tree* parent,
    FILE*                           fp,
    const char*                     model_dirname
)
{
    uint32_t s, num_subtrees;
    uint32_t h, num_haplos; 
    uint32_t parent_haplo;
    char     haplo_buf[256] = {0};
    char     model_fname[1024] = {0};

    Berno_mixture_model_tree*  subtree;
    Berno_mixture_model_tree** subtrees;
    Berno_mixture_model*       model  = NULL;
    Vector_u32*                haplos = NULL;
    String*                    str = NULL;
    Error*                     e;


    if (*subtree_out != NULL)
    {
        free_berno_mixture_model_tree(*subtree_out);
        *subtree_out = NULL;
    }

    assert(*subtree_out = malloc(sizeof(Berno_mixture_model_tree)));
    subtree = *subtree_out;


    if (!skip_fp_spaces_and_comments(fp))
    {
        return create_error(ERROR_IO, __FILE__, "read_berno_mixture_model_tree",
                __LINE__, "Missing model subtree");
    }

    /* Read the parent haplo type. */
    if (fscanf(fp, "%255s", haplo_buf) != 1)
    {
        return create_error(ERROR_INV_ARG, __FILE__, 
                "read_berno_mixture_model_subtree", __LINE__, 
                "Missing parent haplogroup for subtree");
    }
    if ((e = lookup_haplo_group_index_from_label(&parent_haplo, 
                    (const char*)haplo_buf)) != NULL)
    {
        return e;
    }


    /* Read the number of haplos in the root node and allocate space for them.*/
    if (fscanf(fp, "%u", &num_haplos) != 1 || num_haplos == 0)
    {
        return create_error(ERROR_INV_ARG, __FILE__, 
                "read_berno_mixture_model_subtree", __LINE__, 
                "Invalid number of haplos in model subtree");
    }
    create_vector_u32(&haplos, num_haplos);


    /* Read the number of subtrees in the node and allocate space for them. */
    if (fscanf(fp, "%u", &num_subtrees) != 1)
    {
        return create_error(ERROR_INV_ARG, __FILE__, 
                "read_berno_mixture_model_subtree", __LINE__, 
                "Invalid number of subtrees in model subtree");
    }
    subtrees = malloc(num_subtrees*sizeof(struct Berno_mixture_model_tree*));
    assert(subtrees);


    /* Read each haplo name and the model name. */
    if (!skip_fp_spaces_and_comments(fp))
    {
        return create_error(ERROR_IO, __FILE__,
                "read_berno_mixture_model_subtree", __LINE__, 
                "Missing haplogroup in model subtree");
    }
    for(h = 0; h < num_haplos; h++)
    {
        if (fscanf(fp, "%255s", haplo_buf) != 1)
        {
            return create_error(ERROR_INV_ARG, __FILE__, 
                    "read_berno_mixture_model_subtree", __LINE__, 
                    "Missing haplogroup in model subtree");
        }
        if ((e = lookup_haplo_group_index_from_label(&(haplos->elts[ h ]), 
                        (const char*)haplo_buf)) != NULL)
        {
            return e;
        }
    }
    if (fscanf(fp, "%255s", haplo_buf) != 1)
    {
        return create_error(ERROR_INV_ARG, __FILE__, 
                "read_berno_mixture_model_subtree", __LINE__, 
                "Missing model file name in model subtree");
    }
    if (model_dirname)
    {
        snprintf(model_fname, 1023, "%s/%s", model_dirname, haplo_buf);
    }
    else
    {
        snprintf(model_fname, 1023, "%s", haplo_buf);
    }
    if ((e = read_berno_mixture_model(&model, model_fname)) != NULL)
    {
        string_printf(&str, 256, "%s: %s", model_fname, e->msg);
        return create_error(ERROR_INV_ARG, __FILE__, 
                "read_berno_mixture_model_subtree", __LINE__, str->chars);
    }


    subtree->parent_haplo = parent_haplo;
    subtree->parent       = parent;
    subtree->num_subtrees = num_subtrees;
    subtree->subtrees     = subtrees;
    subtree->haplos       = haplos;
    subtree->model        = model;

    /* Read the subtrees. */
    for(s = 0; s < num_subtrees; s++)
    {
        subtrees[ s ] = NULL;
        if ((e = read_berno_mixture_model_subtree(&(subtrees[ s ]), subtree, fp,
                        model_dirname)) != NULL)
        {
            return e;
        }
    }

    return NULL;
}


/**
 * @param  tree_out       Result parameter. If @em *tree_out is NULL, a model 
 *                        tree is allocated; otherwise it is first freed.
 * @param  tree_fname     File to read the model tree from.
 * @param  model_dirname  Directory to read the models from. If NULL, the
 *                        current working directory is used.
 *
 * @return On success, NULL is returned; otherwise an error is returned, but
 * the result parameters are not freed.
 */
Error* read_berno_mixture_model_tree
(
    Berno_mixture_model_tree** tree_out,
    const char*                tree_fname,
    const char*                model_dirname
)
{
    uint32_t s, num_subtrees;
    uint32_t h, num_haplos; 
    FILE*    fp;
    char     c;
    char     haplo_buf[256] = {0};
    char     model_fname[1024] = {0};

    Berno_mixture_model_tree*  tree;
    Berno_mixture_model_tree** subtrees;
    Berno_mixture_model*       model  = NULL;
    Vector_u32*                haplos = NULL;
    String*                    str = NULL;
    Error*                     e;


    if (*tree_out != NULL)
    {
        free_berno_mixture_model_tree(*tree_out);
        *tree_out = NULL;
    }

    if ((fp = fopen(tree_fname, "r")) == NULL)
    {
        return create_error(ERROR_IO, __FILE__, "read_berno_mixture_model_tree",
                __LINE__, strerror(errno));
    }

    assert(*tree_out = malloc(sizeof(Berno_mixture_model_tree)));
    tree = *tree_out;


    if (!skip_fp_spaces_and_comments(fp))
    {
        return create_error(ERROR_IO, __FILE__, "read_berno_mixture_model_tree",
                __LINE__, "Nothing to read in model tree");
    }

    if ((c = fgetc(fp)) != '*')
    {
        return create_error(ERROR_IO, __FILE__, "read_berno_mixture_model_tree",
                __LINE__, "Missing root in model tree");
    }


    /* Read the number of haplos in the root node and allocate space for them.*/
    if (fscanf(fp, "%u", &num_haplos) != 1 || num_haplos == 0)
    {
        return create_error(ERROR_INV_ARG, __FILE__,
                "read_berno_mixture_model_tree", __LINE__, 
                "Invalid number of haplos in model tree root");
    }
    create_vector_u32(&haplos, num_haplos);


    /* Read the number of subtrees in the root node and allocate space for
     * them. */
    if (fscanf(fp, "%u", &num_subtrees) != 1)
    {
        return create_error(ERROR_INV_ARG, __FILE__, 
                "read_berno_mixture_model_tree", __LINE__, 
                "Invalid number of subtrees in model tree root");
    }
    subtrees = malloc(num_subtrees*sizeof(struct Berno_mixture_model_tree*));
    assert(subtrees);


    /* Read each haplo name and the model name. */
    if (!skip_fp_spaces_and_comments(fp))
    {
        return create_error(ERROR_INV_ARG, __FILE__, 
                "read_berno_mixture_model_tree", __LINE__, 
                "Missing haplogroup in model tree");
    }
    for(h = 0; h < num_haplos; h++)
    {
        if (fscanf(fp, "%255s", haplo_buf) != 1)
        {
            return create_error(ERROR_INV_ARG, __FILE__, 
                    "read_berno_mixture_model_tree", __LINE__, 
                    "Missing haplogroup in model tree root");
        }
        if ((e = lookup_haplo_group_index_from_label(&(haplos->elts[ h ]), 
                        (const char*)haplo_buf)) != NULL)
        {
            return e;
        }
    }
    if (fscanf(fp, "%255s", haplo_buf) != 1)
    {
        return create_error(ERROR_INV_ARG, __FILE__, 
                "read_berno_mixture_model_tree", __LINE__, 
                "Missing model file name in model tree root");
    }
    if (model_dirname)
    {
        snprintf(model_fname, 1023, "%s/%s", model_dirname, haplo_buf);
    }
    else
    {
        snprintf(model_fname, 1023, "%s", haplo_buf);
    }
    if ((e = read_berno_mixture_model(&model, model_fname)) != NULL)
    {
        string_printf(&str, 256, "%s: %s", model_fname, e->msg);
        return create_error(ERROR_INV_ARG, __FILE__,
                "read_berno_mixture_model_tree", __LINE__, str->chars); 
    }


    tree->parent_haplo = 0;
    tree->parent       = NULL;
    tree->num_subtrees = num_subtrees;
    tree->subtrees     = subtrees;
    tree->haplos       = haplos;
    tree->model        = model;

    /* Read the subtrees. */
    for(s = 0; s < num_subtrees; s++)
    {
        subtrees[ s ] = NULL;
        if ((e = read_berno_mixture_model_subtree(&(subtrees[ s ]), tree, fp,
                        model_dirname)) != NULL)
        {
            return e;
        }
    }

    if (fclose(fp) != 0)
    {
        return create_error(ERROR_IO, __FILE__, "read_berno_mixture_model_tree",
                __LINE__, strerror(errno));
    }

    return NULL;
}





/** @param  tree  Model tree to free. */
void free_berno_mixture_model_tree(Berno_mixture_model_tree* tree)
{
    uint32_t s;

    if (!tree)
        return;

    for(s = 0; s < tree->num_subtrees; s++)
    {
        free_berno_mixture_model_tree(tree->subtrees[ s ]);
    }

    free(tree->subtrees);
    free_vector_u32(tree->haplos);
    free_berno_mixture_model(tree->model);
    free(tree);
}




/** 
 * @brief Tests a predicted haplogroup against an actual haplogroup name based
 * on substring prefixes.
 */
static Berno_mixture_accuracy test_haplogroup_substrings
(
    uint32_t pred_haplo,
    uint32_t actual_haplo
)
{
    const char* pred_haplo_name;
    const char* actual_haplo_name;

    assert(lookup_haplo_group_label_from_index(&pred_haplo_name, pred_haplo) 
            == NULL);
    assert(lookup_haplo_group_label_from_index(&actual_haplo_name, actual_haplo)
            == NULL);

    if (strcmp(pred_haplo_name, actual_haplo_name) == 0)
    {
        return BERNO_MIX_CORRECT;
    }
    if (strncmp(pred_haplo_name, actual_haplo_name, strlen(pred_haplo_name)) 
            == 0)
    {
        return BERNO_MIX_UP_TREE;
    }
    if (strncmp(pred_haplo_name, actual_haplo_name, strlen(actual_haplo_name)) 
            == 0)
    {
        return BERNO_MIX_DOWN_TREE;
    }

    return BERNO_MIX_INCORRECT;
}


/** @brief Recursively searches a model tree for a haplogroup. */
static const Berno_mixture_model_tree* recursively_find_haplo
(
    uint32_t                        haplo,
    const Berno_mixture_model_tree* tree
)
{
    uint32_t h, num_haplos;
    uint32_t s, num_subtrees;
    Berno_mixture_accuracy accuracy;
    const Berno_mixture_model_tree* subtree;

    num_haplos   = tree->haplos->num_elts;
    num_subtrees = tree->num_subtrees;

    for (h = 0; h < num_haplos; h++)
    {
        accuracy = test_haplogroup_substrings(haplo, tree->haplos->elts[ h ]);
        if (accuracy == BERNO_MIX_CORRECT || accuracy == BERNO_MIX_UP_TREE)
        {
            return tree;
        }
        if (accuracy == BERNO_MIX_DOWN_TREE)
        {
            for (s = 0; s < num_subtrees; s++)
            {
                if (tree->subtrees[s]->parent_haplo == tree->haplos->elts[h])
                {
                    if ((subtree = recursively_find_haplo(haplo, 
                                tree->subtrees[ s ])) != NULL)
                    {
                        return subtree;
                    }
                }
            }
            return tree;
        }
    }

    for (s = 0; s < num_subtrees; s++)
    {
        if ((subtree = recursively_find_haplo(haplo, tree->subtrees[ s ])) 
                != NULL)
        {
            return subtree;
        }
    }

    return NULL;
}


/** 
 * @brief Returns whether or not @em parent is the parent tree of 
 * @em subtree.
 */
static uint8_t is_parent_of_subtree
(
    uint32_t                        parent_haplo,
    const Berno_mixture_model_tree* parent,
    const Berno_mixture_model_tree* subtree
)
{
    uint32_t num_haplos;
    uint32_t h;
    uint32_t haplo;

    if (parent == subtree)
    {
        return 0;
    }

    num_haplos = parent->haplos->num_elts;

    for (h = 0; h < num_haplos; h++)
    {
        haplo = parent->haplos->elts[ h ];
        if (subtree->parent_haplo == haplo &&
            test_haplogroup_substrings(parent_haplo, haplo) 
                 != BERNO_MIX_INCORRECT)
        {
            return 1;
        }
    }
    if (subtree->parent != NULL)
    {
        return is_parent_of_subtree(parent_haplo, parent, subtree->parent);
    }
    return 0;
}


/** 
 * @param  pred_haplo    Predicted haplogroup.
 * @param  actual_haplo  Actual haplogroup.
 * @param  tree          Model tree.
 *
 * @return Returns the accuracy of the prediction.
 */
Berno_mixture_accuracy berno_mixture_model_tree_haplo_accuracy
(
    uint32_t                        pred_haplo,
    uint32_t                        actual_haplo,
    const Berno_mixture_model_tree* tree
)
{
    const Berno_mixture_model_tree* pred_subtree;
    const Berno_mixture_model_tree* actual_subtree;
    Berno_mixture_accuracy result;

    if ((result = test_haplogroup_substrings(pred_haplo, actual_haplo)) 
            != BERNO_MIX_INCORRECT)
    {
        return result;
    }

    assert(pred_subtree = recursively_find_haplo(pred_haplo, tree));
    assert(actual_subtree = recursively_find_haplo(actual_haplo, tree));

    if (is_parent_of_subtree(pred_haplo, pred_subtree, actual_subtree))
    {
        result = BERNO_MIX_UP_TREE;
    }
    else if (is_parent_of_subtree(actual_haplo, actual_subtree, pred_subtree))
    {
        result = BERNO_MIX_DOWN_TREE;
    }

    return result;
}
