/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Bernoulli mixture model classifier.
 */


#ifndef HAPLO_BMM_H
#define HAPLO_BMM_H


#include <config.h>

#include <stdlib.h>
#include <inttypes.h>

#include <jwsc/base/error.h>
#include <jwsc/vector/vector.h>
#include <jwsc/matrix/matrix.h>


/** 
 * @brief Bernoulli mixture model.
 *
 * A BMM is created conditioned on each haplogroup; the BMM parameters (mu,
 * sigma and pi) are actually in arrays indexed by haplogroup.
 */
typedef struct
{
    /** @brief Number of markers. */
    uint32_t  num_markers;

    /** @brief Number of haplogroups. */
    uint32_t  num_haplogroups;

    /** @brief Number of mixture components. */
    uint32_t  num_mix;

    /** 
     * @brief Parameters for the mixture of Bernoullis over the markers for each
     * haplogroup.
     */
    Matrix_d** mu;

    /** @brief Mixture component weights. */
    Vector_d** pi;

    /** @brief Prior probabilities over the haplogroups. */
    Vector_d* haplo_priors;
} 
Berno_mixture_model;


/** @brief Bernoulli mixture model tree. */
typedef struct Berno_mixture_model_tree
{
    /** @brief Parent haplogroup name. */
    uint32_t parent_haplo;

    /** @brief Pointer to the parent haplogroup model tree. */
    const struct Berno_mixture_model_tree*  parent;

    /** @brief Number of model subtrees. */
    uint32_t  num_subtrees;

    /** @brief Array of model subtrees pointers. */
    struct Berno_mixture_model_tree** subtrees;

    /** @brief Haplogroups in this model tree node. */
    Vector_u32* haplos;

    /** @brief Bernoulli mixture model. */
    Berno_mixture_model* model;
} 
Berno_mixture_model_tree;


/** @brief Haplogroup prediction accuracy for a Bernoulli mixture model. */
typedef enum 
{
    BERNO_MIX_INCORRECT,
    BERNO_MIX_CORRECT,
    BERNO_MIX_UP_TREE,
    BERNO_MIX_DOWN_TREE
}
Berno_mixture_accuracy;




/** @brief Trains a Bernoulli mixture model. */
void train_berno_mixture_model
(
    Berno_mixture_model**  model_out,
    const Vector_u32*          haplos, 
    const Matrix_i32*          markers,
    const Vector_d*            haplo_priors,
    uint32_t                   num_mix
);


/** 
 * @brief Trains a set of Bernoulli mixture models in a model information file. 
 */
Error* train_berno_mixture_model_info
(
    const Vector_u32* haplos, 
    const Matrix_i32* markers,
    const char*       info_fname,
    const char*       model_dirname
);


/** @brief Writes a Bernoulli mixture model. */
Error* write_berno_mixture_model(Berno_mixture_model* model, const char* fname);


/** @brief Reads a Bernoulli mixture model. */
Error* read_berno_mixture_model(Berno_mixture_model** model_out, const char* fname);


/** @brief Frees a Bernoulli mixture model. */
void free_berno_mixture_model(Berno_mixture_model* model);


/** 
 * @brief Predicts a haplogroup for a marker sample using a Bernoulli mixture
 * model.
 */
Error* predict_haplo_with_berno_mixture_model
(
    uint32_t*                  haplo_out,
    double*                    confidence_out,
    const Vector_i32*          markers,
    const Berno_mixture_model* model
);


/** 
 * @brief Predicts the haplogroups for a set of marker samples using a
 * Bernoulli mixture model.
 */
Error* predict_haplos_with_berno_mixture_model
(
    Vector_u32**               haplos_out,
    Vector_d**                 confidence_out,
    const Matrix_i32*          markers,
    const Berno_mixture_model* model
);


/** 
 * @brief Predicts the haplogroups for a set of marker samples using a 
 * Bernoulli mixture model tree.
 */
Error* predict_haplos_with_berno_mixture_model_tree
(
    Vector_u32**                    haplos_out,
    Vector_d**                      confidence_out,
    const Matrix_i32*               markers,
    const Berno_mixture_model_tree* tree
);


/** @brief Reads a Bernoulli mixture model tree from. */
Error* read_berno_mixture_model_tree
(
    Berno_mixture_model_tree** tree_out,
    const char*                tree_fname,
    const char*                model_dirname
);


/** @brief Frees a Bernoulli mixture model tree. */
void free_berno_mixture_model_tree(Berno_mixture_model_tree* tree);


/** 
 * @brief Tests the accuracy of a predicted haplogroup under a Bernoulli
 * mixture model tree.
 */
Berno_mixture_accuracy berno_mixture_model_tree_haplo_accuracy
(
    uint32_t                        pred_haplo,
    uint32_t                        actual_haplo,
    const Berno_mixture_model_tree* tree
);


#endif
