/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Gaussian and Bernoulli mixture model classifier.
 */


#include <config.h>

#include <stdlib.h>
#include <stdio.h>
#include <inttypes.h>
#include <assert.h>
#include <string.h>
#include <errno.h>
#include <math.h>

#include <jwsc/base/error.h>
#include <jwsc/base/limits.h>
#include <jwsc/base/string.h>
#include <jwsc/base/file_io.h>
#include <jwsc/prob/pdf.h>
#include <jwsc/vector/vector.h>
#include <jwsc/vector/vector_io.h>
#include <jwsc/vector/vector_math.h>
#include <jwsc/matrix/matrix.h>
#include <jwsc/matrix/matrix_math.h>
#include <jwsc/matrix/matrix_io.h>
#include <jwsc/matblock/matblock.h>
#include <jwsc/matblock/matblock_io.h>
#include <jwsc/stat/stat.h>
#include <jwsc/stat/gmm.h>
#include <jwsc/stat/bmm.h>

#include "haplo_groups.h"
#include "gmm_bmm.h"


/**
 * @param  model_haplos_out        Haplogroups for the model in the info entry.
 * @param  model_markers_out       Markers for the model in the info entry.
 * @param  model_priors_out        Priors over the haplogroups in the info
 *                                 entry. Only use if @em
 *                                 *model_priors_given_out == 1.
 * @param  model_priors_given_out  Result parameter. Boolean value indicates 
 *                                 whether or not the @em model_priors_out 
 *                                 result parameter was initialized from the
 *                                 info file and should be used.
 * @param  model_fname_out         Name of the model for info file entry read.
 * @param  num_mix_out             Number of mixture components.
 * @param  haplos                  Haplogroups to use for constructing the 
 *                                 subset used for the model in the info entry.
 * @param  markers                 Markers to use for constructing the 
 *                                 subset used for the model in the info entry.
 * @param  fp                      File stream to read the entry from.
 *
 * @return On success, NULL is returned; otherwise NULL is returned but does not
 * free result parameters. 
 */
static Error* read_model_info_entry
(
    Vector_u32**      model_haplos_out,
    Matrix_i32**      model_markers_out,
    uint32_t*         num_gmm_mix_out,
    uint32_t*         num_bmm_mix_out,
    Vector_d**        model_priors_out,
    uint8_t*          model_priors_given_out,
    String**          model_gmm_fname_out,
    String**          model_bmm_fname_out,
    const Vector_u32* haplos,
    const Matrix_i32* markers,
    FILE*             fp
)
{
    uint32_t num_haplogroups;
    uint32_t num_haplos;
    uint32_t h;
    uint32_t num_aliases;
    uint32_t a;
    uint32_t num_samples;
    uint32_t sample;
    uint32_t num_samples_out;
    uint32_t num_markers;
    uint32_t marker;
    uint32_t haplo;
    uint32_t label;
    uint32_t alias;
    uint32_t model_priors_given;
    uint8_t  found;
    char     haplo_buf[256] = {0};
    char     aliases_buf[1024] = {0};
    char     gmm_fname_buf[256] = {0};
    char     bmm_fname_buf[256] = {0};
    char*    buf_ptr;

    const char* alias_name;
    const char* haplo_name;

    Vector_u32* model_haplos;
    Matrix_i32* model_markers;
    Vector_d*   model_priors;
    String*     model_gmm_fname;
    String*     model_bmm_fname;
    String*     str = NULL;
    Error*      e;

    Vector_u32*  haplo_alias_labels = NULL;
    Vector_u32** haplo_alias        = NULL;

    num_haplogroups = get_num_haplo_groups();

    if (fscanf(fp, "%255s %255s %u %u %u %u\n", gmm_fname_buf, bmm_fname_buf,
                &num_haplos, num_gmm_mix_out, num_bmm_mix_out,
                &model_priors_given) != 6)
    {
        return create_error(ERROR_INV_ARG, __FILE__, "read_model_info_entry",
                __LINE__, "Model information file not formatted properly");
    }
    *model_priors_given_out = (model_priors) ? 1 : 0;

    create_string_from_char(model_gmm_fname_out, gmm_fname_buf);
    model_gmm_fname = *model_gmm_fname_out;

    create_string_from_char(model_bmm_fname_out, bmm_fname_buf);
    model_bmm_fname = *model_bmm_fname_out;

    num_samples = markers->num_rows;
    num_markers = markers->num_cols;

    create_vector_u32(model_haplos_out, num_samples);
    model_haplos = *model_haplos_out;

    create_matrix_i32(model_markers_out, num_samples, num_markers);
    model_markers = *model_markers_out;

    if (model_priors_given)
    {
        create_vector_d(model_priors_out, num_haplogroups);
        model_priors = *model_priors_out;
    }

    create_vector_u32(&haplo_alias_labels, num_haplos);
    assert(haplo_alias = malloc(num_haplos*sizeof(Vector_u32)));

    for (h = 0; h < num_haplos; h++)
    {
        if (fscanf(fp, "%255s", haplo_buf) != 1)
        {
            return create_error(ERROR_INV_ARG, __FILE__, 
                    "read_model_info_entry", __LINE__, 
                    "Model information file not formatted properly");
        }

        if ((e = lookup_haplo_group_index_from_label(
                        &(haplo_alias_labels->elts[ h ]),
                        (const char*)haplo_buf)) != NULL)
        {
            return e;
        }

        if (fscanf(fp, "%1023s", aliases_buf) != 1)
        {
            return create_error(ERROR_INV_ARG, __FILE__, 
                    "read_model_info_entry", __LINE__, 
                    "Model information file not formatted properly");
        }

        num_aliases = 1;
        buf_ptr = aliases_buf;
        while (*buf_ptr != '\0')
        {
            if (*buf_ptr == ':')
            {
                *buf_ptr = '\0';
                num_aliases++;
            }
            buf_ptr++;
        }

        haplo_alias[ h ] = NULL;
        create_vector_u32(&(haplo_alias[ h ]), num_aliases);

        buf_ptr = aliases_buf;
        for (a = 0; a < num_aliases; a++)
        {
            assert(sscanf(buf_ptr, "%255s", haplo_buf) == 1);
            if ((e = lookup_haplo_group_index_from_label(
                            &(haplo_alias[ h ]->elts[ a ]),
                            (const char*)haplo_buf)) != NULL)
            {
                return e;
            }

            while (*buf_ptr != '\0')
            {
                buf_ptr++;
            }
            buf_ptr++;
        }

        if (model_priors_given)
        {
            label = haplo_alias_labels->elts[ h ];
            if (fscanf(fp, "%lf", &(model_priors->elts[ label ])) != 1)
            {
                return create_error(ERROR_INV_ARG, __FILE__, 
                        "read_model_info_entry", __LINE__, 
                        "Model information file not formatted properly");
            }
        }
    }


    num_samples_out = 0;

    for (sample = 0; sample < num_samples; sample++)
    {
        haplo = haplos->elts[ sample ];
        found = 0;

        for (h = 0; !found && h < num_haplos; h++)
        {
            label = haplo_alias_labels->elts[ h ];
            num_aliases = haplo_alias[ h ]->num_elts;

            for (a = 0; !found && a < num_aliases; a++)
            {
                alias = haplo_alias[ h ]->elts[ a ];

                assert(lookup_haplo_group_label_from_index(&alias_name, alias) 
                        == NULL);
                assert(lookup_haplo_group_label_from_index(&haplo_name, haplo) 
                        == NULL);

                if (strncmp(alias_name, haplo_name, strlen(alias_name)) == 0)
                {
                    found = 1;
                }
            }
        }

        if (found)
        {
            model_haplos->elts[ num_samples_out ] = label;
            for (marker = 0; marker < num_markers; marker++)
            {
                model_markers->elts[ num_samples_out ][ marker ] =
                    markers->elts[ sample ][ marker ];
            }
            num_samples_out++;
        }
    }


    if (num_samples_out == 0)
    {
        string_printf(&str, 256, "No haplo samples to train model '%s'", 
                model_gmm_fname->chars);
        return create_error(ERROR_INV_ARG, __FILE__, "read_model_info_entry", 
                __LINE__, str->chars);
    }


    *model_haplos_out = NULL;
    assert(copy_vector_section_u32(model_haplos_out, model_haplos, 0, 
                num_samples_out) == NULL);
    free_vector_u32(model_haplos);

    *model_markers_out = NULL;
    assert(copy_matrix_block_i32(model_markers_out, model_markers, 0, 0,
                num_samples_out, num_markers) == NULL);
    free_matrix_i32(model_markers);


    free_vector_u32(haplo_alias_labels);
    for (h = 0; h < num_haplos; h++)
    {
        free_vector_u32(haplo_alias[ h ]);
    }
    free(haplo_alias);

    return NULL;
}


void split_gmm_bmm_markers
(
    Matrix_i32**      gmm_markers_out,
    Matrix_i32**      bmm_markers_out,
    const Matrix_i32* markers,
    uint32_t          num_gmm_markers
)
{
    uint32_t num_samples;
    uint32_t num_markers;

    num_samples = markers->num_rows;
    num_markers = markers->num_cols;

    copy_matrix_block_i32(gmm_markers_out, markers, 0, 0, num_samples,
            num_gmm_markers);

    copy_matrix_block_i32(bmm_markers_out, markers, 0, num_gmm_markers, 
            num_samples, num_markers - num_gmm_markers);
}



/**
 * @param  model_out     Result parameter. If @em *model_out is NULL, a
 *                       model is allocated; otherwise its space is re-used.
 * @param  haplos        Haplogroups for training, with ith element as 
 *                       corresponding to the ith sample in @em markers.
 * @param  markers       Markers for training, with ith row as a sample 
 *                       corresponding to the ith haplo in @em haplos.
 * @param  haplo_priors  Priors for each of the haplogroups.
 * @param  num_mix       Number of mixture components.
 */
void train_gauss_berno_mixture_model
(
    Gauss_berno_mixture_model** model_out,
    const Vector_u32*           haplos, 
    const Matrix_i32*           markers,
    uint32_t                    num_gmm_markers,
    uint32_t                    num_gmm_mix,
    uint32_t                    num_bmm_mix,
    const Vector_d*             haplo_priors
)
{
    Matrix_i32* gmm_markers = NULL;
    Matrix_i32* bmm_markers = NULL;

    split_gmm_bmm_markers(&gmm_markers, &bmm_markers, markers, num_gmm_markers);

    if (*model_out == NULL)
    {
        assert(*model_out = malloc(sizeof(Gauss_berno_mixture_model)));
        (*model_out)->gmm = NULL;
        (*model_out)->bmm = NULL;
    }

    train_gauss_mixture_model(&((*model_out)->gmm), haplos, gmm_markers,
            haplo_priors, num_gmm_mix);

    train_berno_mixture_model(&((*model_out)->bmm), haplos, bmm_markers,
            haplo_priors, num_bmm_mix);

    free_matrix_i32(gmm_markers);
    free_matrix_i32(bmm_markers);
}




/**
 * The models are written to the file names specified in the model information
 * file.
 *
 * @param  haplos         Haplogroups for training, with ith element as 
 *                        corresponding to the ith sample in @em markers.
 * @param  markers        Markers for training, with ith row as a sample 
 *                        corresponding to the ith haplo in @em haplos.
 * @param  info_fname     File containing the model information.
 * @param  model_dirname  Directory to write the models to. If NULL, the current
 *                        working directory is used.
 *
 * @return On success, NULL is returned; otherwise an error is returned.
 */
Error* train_gauss_berno_mixture_model_info
(
    const Vector_u32* haplos, 
    const Matrix_i32* markers,
    uint32_t          num_gmm_markers,
    const char*       info_fname,
    const char*       model_dirname
)
{
    FILE* fp;

    Gauss_berno_mixture_model* model;
    uint32_t             num_samples;
    uint32_t             num_markers;
    uint32_t             num_gmm_mix;
    uint32_t             num_bmm_mix;

    Vector_u32*     model_haplos;
    Matrix_i32*     model_markers;
    Vector_d*       model_priors;
    uint8_t         model_priors_given;
    String*         model_gmm_fname;
    String*         model_bmm_fname;
    String*         str;
    Error*          e;
    char            fname_buf[1024];


    model             = NULL;
    model_haplos      = NULL;
    model_markers     = NULL;
    model_priors      = NULL;
    model_gmm_fname   = NULL;
    model_bmm_fname   = NULL;
    str               = NULL;

    num_samples = markers->num_rows;
    num_markers = markers->num_cols;

    if ((fp = fopen(info_fname, "r")) == NULL)
    {
        return create_error(ERROR_IO, __FILE__, 
                "train_gauss_berno_mixture_model_info", __LINE__,
                strerror(errno));
    }

    while (skip_fp_spaces_and_comments(fp))
    {
        model_priors_given = 0;

        if ((e = read_model_info_entry(&model_haplos, &model_markers,
                        &num_gmm_mix, &num_bmm_mix, &model_priors, 
                        &model_priors_given, &model_gmm_fname,
                        &model_bmm_fname, haplos, markers, fp)) != NULL)
        {
            return e;
        }

        train_gauss_berno_mixture_model(&model, model_haplos,
                model_markers, num_gmm_markers, num_gmm_mix, num_bmm_mix,
                model_priors_given ?  model_priors : NULL);

        if (model_dirname)
        {
            snprintf(fname_buf, 1023, "%s/%s", model_dirname, 
                    model_gmm_fname->chars);
        }
        else
        {
            snprintf(fname_buf, 1023, "%s", model_gmm_fname->chars);
        }

        if ((e = write_gauss_mixture_model(model->gmm, fname_buf)) != NULL)
        {
            string_printf(&str, 256, "%s: %s", model_gmm_fname->chars, e->msg);
            e = create_error(ERROR_IO, __FILE__, 
                    "train_gauss_berno_mixture_model_info", __LINE__,
                    str->chars);
            return e;
        }

        if (model_dirname)
        {
            snprintf(fname_buf, 1023, "%s/%s", model_dirname, 
                    model_bmm_fname->chars);
        }
        else
        {
            snprintf(fname_buf, 1023, "%s", model_bmm_fname->chars);
        }

        if ((e = write_berno_mixture_model(model->bmm, fname_buf)) != NULL)
        {
            string_printf(&str, 256, "%s: %s", model_bmm_fname->chars, e->msg);
            e = create_error(ERROR_IO, __FILE__, 
                    "train_gauss_berno_mixture_model_info", __LINE__,
                    str->chars);
            return e;
        }
    }

    free_gauss_berno_mixture_model(model);
    free_vector_u32(model_haplos);
    free_matrix_i32(model_markers);
    free_vector_d(model_priors);
    free_string(model_gmm_fname);
    free_string(model_bmm_fname);

    if (fclose(fp) != 0)
    {
        return create_error(ERROR_IO, __FILE__, 
                "train_gauss_berno_mixture_model_info", __LINE__,
                strerror(errno));
    }

    return NULL;
}




/**
 * @param  model  Model to write.
 * @param  fname  File to write to.
 *
 * @return On success, NULL is returned; otherwise an error is returned.
 */
Error* write_gauss_berno_mixture_model
(
    Gauss_berno_mixture_model* model, 
    const char* gmm_fname,
    const char* bmm_fname
)
{
    Error* e;

    assert(model->gmm->num_markers == model->bmm->num_markers);
    assert(model->gmm->num_haplogroups == model->bmm->num_haplogroups);

    if ((e = write_gauss_mixture_model(model->gmm, gmm_fname)) ||
        (e = write_berno_mixture_model(model->bmm, bmm_fname)))
    {
        return e;
    }

    return NULL;
}




/**
 * @param  model_out  Result parameter. If @em *model_out is NULL, a model is 
 *                    allocated; otherwise its space is re-used.
 * @param  fname      File to read the model from.
 *
 * @return On success, NULL is returned; otherwise an error is returned but does
 * not free result parameter.
 */
Error* read_gauss_berno_mixture_model
(
    Gauss_berno_mixture_model** model_out, 
    const char*           gmm_fname,
    const char*           bmm_fname
)
{
    Error* e;

    if(*model_out == NULL)
    {
        assert((*model_out = malloc(sizeof(Gauss_berno_mixture_model))));
        (*model_out)->gmm = NULL;
        (*model_out)->bmm = NULL;
    }

    if ((e = read_gauss_mixture_model(&((*model_out)->gmm), gmm_fname)) ||
        (e = read_berno_mixture_model(&((*model_out)->bmm), bmm_fname)))
    {
        return e;
    }

    return NULL;
}




/** @param  model  Model to free. */
void free_gauss_berno_mixture_model(Gauss_berno_mixture_model* model)
{
    if (!model)
        return;

    free_gauss_mixture_model(model->gmm);
    free_berno_mixture_model(model->bmm);
    free(model);
}




/** 
 * @param  haplo_out       Result parameter. 
 * @param  confidence_out  Result parameter. 
 * @param  markers         Marker data to predict.
 * @param  model           Trained model to use for predicting.
 *
 * @return On success, NULL is returned; otherwise an error is returned.
 */
Error* predict_haplo_with_gauss_berno_mixture_model
(
    uint32_t*                  haplo_out,
    double*                    confidence_out,
    const Vector_i32*          gmm_markers,
    const Vector_i32*          bmm_markers,
    const Gauss_berno_mixture_model* model
)
{
    uint32_t m, num_gmm_markers, num_bmm_markers;
    uint32_t h, best_h, num_haplogroups;
    uint32_t i;

    double log_prior;
    double log_map;
    double log_ll;
    double log_posterior;
    double posterior_sum;

    Matrix_d*   markers_d   = NULL;
    Matrix_u32* markers_u32 = NULL;

    num_haplogroups = get_num_haplo_groups();
    num_gmm_markers = gmm_markers->num_elts;
    num_bmm_markers = bmm_markers->num_elts;

    assert(num_haplogroups == model->gmm->num_haplogroups);
    assert(num_gmm_markers == model->gmm->num_markers);
    assert(num_bmm_markers == model->bmm->num_markers);

    log_map = log(0);
    posterior_sum = 0;

    create_matrix_d(&markers_d, 1, num_gmm_markers);
    create_matrix_u32(&markers_u32, 1, num_bmm_markers);

    for(h = 0; h < num_haplogroups; h++)
    {
        if (!model->gmm->mu[ h ])
        {
            assert(!model->bmm->mu[ h ]);
            continue;
        }

        for (m = 0; m < num_gmm_markers; m++)
        {
            if (gmm_markers->elts[ m ] <= 0)
            {
                // Impute missing value with the mean.
                markers_d->elts[0][ m ] = 0;
                for (i = 0; i < model->gmm->num_mix; i++)
                {
                    markers_d->elts[0][ m ] += 
                        model->gmm->mu[ h ]->elts[ i ][ m ] *
                        model->gmm->pi[ h ]->elts[ i ];
                }
            }
            else
            {
                markers_d->elts[0][ m ] = gmm_markers->elts[ m ];
            }
        }

        log_ll = gmm_log_likelihood_d(model->gmm->mu[ h ], 
                model->gmm->sigma[ h ], model->gmm->pi[ h ], markers_d);

        for (m = 0; m < num_bmm_markers; m++)
        {
            if (bmm_markers->elts[ m ] < 0)
            {
                // Impute missing value with the mean.
                // TODO not correct computation here
                markers_u32->elts[0][ m ] = 0;
                for (i = 0; i < model->bmm->num_mix; i++)
                {
                    markers_u32->elts[0][ m ] += 
                        model->bmm->mu[ h ]->elts[ i ][ m ] *
                        model->bmm->pi[ h ]->elts[ i ];
                }
            }
            else
            {
                markers_u32->elts[0][ m ] = bmm_markers->elts[ m ];
            }
        }

        log_ll += bmm_log_likelihood_d(model->bmm->mu[ h ], 
                model->bmm->pi[ h ], markers_u32);

        if (model->gmm->haplo_priors->elts[ h ] < JWSC_MIN_LOG_ARG)
        {
            log_prior = log(JWSC_MIN_LOG_ARG);
        }
        else
        {
            log_prior = log(model->gmm->haplo_priors->elts[ h ]);
        }

        log_posterior = log_ll + log_prior;
        posterior_sum += exp(log_posterior);

        if (log_posterior > log_map)
        {
            log_map = log_posterior;
            best_h = h;
        }
    }

    if (posterior_sum < JWSC_MIN_LOG_ARG)
    {
        *haplo_out = best_h;
        *confidence_out = 0;
    }
    else
    {
        log_map -= log(posterior_sum);

        *haplo_out = best_h;
        *confidence_out = exp(log_map);
    }

    free_matrix_d(markers_d);
    free_matrix_u32(markers_u32);

    return NULL;
}




/** 
 * @param  haplos_out      Result parameter. If @em *haplos_out is NULL, it is
 *                         allocated; otherwise its space is re-used.
 * @param  confidence_out  Result parameter. If @em *confidence_out is NULL, it 
 *                         is allocated; otherwise its space is re-used.
 * @param  markers         Marker data to predict. Each row is a sample for
 *                         prediction, corresponding to an element in the result
 *                         parameters.
 * @param  model           Trained model to use for predicting.
 *
 * @return On success, NULL is returned; otherwise an error is returned, but the
 * result parameters are not freed.
 */
Error* predict_haplos_with_gauss_berno_mixture_model
(
    Vector_u32**              haplos_out,
    Vector_d**                confidence_out,
    const Matrix_i32*         markers,
    uint32_t                  num_gmm_markers,
    const Gauss_berno_mixture_model* model
)
{
    uint32_t s, num_samples;
    uint32_t m, num_markers, num_bmm_markers;

    Vector_u32* haplos;
    Vector_d*   confidence;
    Matrix_i32* gmm_markers   = NULL;
    Matrix_i32* bmm_markers   = NULL;
    Vector_i32* gmm_markers_v = NULL;
    Vector_i32* bmm_markers_v = NULL;
    Error*      e;

    num_samples     = markers->num_rows;
    num_markers     = markers->num_cols;
    num_bmm_markers = num_markers - num_gmm_markers;

    assert(num_gmm_markers == model->gmm->num_markers);
    assert(num_bmm_markers == model->bmm->num_markers);

    split_gmm_bmm_markers(&gmm_markers, &bmm_markers, markers, num_gmm_markers);

    create_vector_u32(haplos_out, num_samples);
    haplos = *haplos_out;

    create_vector_d(confidence_out, num_samples);
    confidence = *confidence_out;

    create_vector_i32(&gmm_markers_v, num_gmm_markers);
    create_vector_i32(&bmm_markers_v, num_bmm_markers);

    for(s = 0; s < num_samples; s++)
    {
        for(m = 0; m < num_gmm_markers; m++)
        {
            gmm_markers_v->elts[ m ] = gmm_markers->elts[ s ][ m ];
        }
        for(m = 0; m < num_bmm_markers; m++)
        {
            bmm_markers_v->elts[ m ] = bmm_markers->elts[ s ][ m ];
        }

        if ((e = predict_haplo_with_gauss_berno_mixture_model(
                        &(haplos->elts[ s ]), &(confidence->elts[ s ]), 
                        gmm_markers_v, bmm_markers_v, model)))
        {
            return e;
        }
    }

    free_matrix_i32(gmm_markers);
    free_matrix_i32(bmm_markers);
    free_vector_i32(gmm_markers_v);
    free_vector_i32(bmm_markers_v);

    return NULL;
}



/** @brief Recursively predicts a haplogroup tree. */
static Error* recursively_predict_haplo_in_model_tree
(
    uint32_t*                       haplo_out,
    double*                         confidence_out,
    const Gauss_berno_mixture_model_tree* tree,
    const Vector_i32*               gmm_markers,
    const Vector_i32*               bmm_markers
)
{
    uint32_t subtree_haplo;
    uint32_t subtree;

    double confidence;

    Error* e;

    if ((e = predict_haplo_with_gauss_berno_mixture_model(haplo_out, 
                    &confidence, gmm_markers, bmm_markers, tree->model)))
    {
        return e;
    }

    *confidence_out *= confidence;

    /* If the best model has a subtree, recursively predict in the subtree. */
    for (subtree = 0; subtree < tree->num_subtrees; subtree++)
    {
        subtree_haplo = tree->subtrees[ subtree ]->parent_haplo;

        if (subtree_haplo == *haplo_out)
        {
            if ((e = recursively_predict_haplo_in_model_tree(haplo_out,
                            confidence_out, tree->subtrees[ subtree ], 
                            gmm_markers, bmm_markers)))
            {
                return e;
            }
        }
    }

    return NULL;
}


/** 
 * @param  haplos_out      Result parameter. If @em *haplos_out is NULL, it is
 *                         allocated; otherwise its space is re-used.
 * @param  confidence_out  Result parameter. If @em *confidence_out is NULL, it 
 *                         is allocated; otherwise its space is re-used.
 * @param  markers         Marker data to predict. Each row is a sample for
 *                         prediction, corresponding to an element in the result
 *                         parameters.
 * @param  treel           Trained model tree to use for predicting.
 *
 * @return On success, NULL is returned; otherwise an error is returned, but the
 * result parameters are not freed.
 */
Error* predict_haplos_with_gauss_berno_mixture_model_tree
(
    Vector_u32**                    haplos_out,
    Vector_d**                      confidence_out,
    const Matrix_i32*               markers,
    uint32_t                        num_gmm_markers,
    const Gauss_berno_mixture_model_tree* tree
)
{
    uint32_t s, num_samples;
    uint32_t m, num_markers, num_bmm_markers;

    Vector_u32* haplos;
    Vector_d*   confidence;
    Matrix_i32* gmm_markers   = NULL;
    Matrix_i32* bmm_markers   = NULL;
    Vector_i32* gmm_markers_v = NULL;
    Vector_i32* bmm_markers_v = NULL;
    Error*      e;

    num_samples     = markers->num_rows;
    num_markers     = markers->num_cols;
    num_bmm_markers = num_markers - num_gmm_markers;

    assert(num_gmm_markers == tree->model->gmm->num_markers);
    assert(num_bmm_markers == tree->model->bmm->num_markers);

    split_gmm_bmm_markers(&gmm_markers, &bmm_markers, markers, num_gmm_markers);

    create_vector_u32(haplos_out, num_samples);
    haplos = *haplos_out;

    create_init_vector_d(confidence_out, num_samples, 1.0);
    confidence = *confidence_out;

    create_vector_i32(&gmm_markers_v, num_gmm_markers);
    create_vector_i32(&bmm_markers_v, num_bmm_markers);

    for (s = 0; s < num_samples; s++ )
    {
        for (m = 0; m < num_gmm_markers; m++ )
        {
            gmm_markers_v->elts[ m ] = gmm_markers->elts[ s ][ m ];
        }
        for (m = 0; m < num_bmm_markers; m++ )
        {
            bmm_markers_v->elts[ m ] = bmm_markers->elts[ s ][ m ];
        }

        if ((e = recursively_predict_haplo_in_model_tree(&(haplos->elts[ s ]), 
                        &(confidence->elts[ s ]), tree, gmm_markers_v,
                        bmm_markers_v)))
        {
            return e;
        }
    }

    free_matrix_i32(gmm_markers);
    free_matrix_i32(bmm_markers);
    free_vector_i32(gmm_markers_v);
    free_vector_i32(bmm_markers_v);

    return NULL;
}




/** 
 * @brief Recursively reads a subtree of a model tree.
 *
 * @param  subtree_out    Result parameter. If @em *tree_out is NULL, a model 
 *                        tree is allocated; otherwise it is first freed.
 * @param  parent         Parent of the subtree being read.
 * @param  fp             File stream to read the model subtree from.
 * @param  model_dirname  Directory to read the models from. If NULL, the
 *                        current working directory is used.
 *
 * @return On success, NULL is returned; otherwise an error is returned, but
 * the result parameters are not freed.
 */
static Error* read_gauss_berno_mixture_model_subtree
(
    Gauss_berno_mixture_model_tree**      subtree_out,
    const Gauss_berno_mixture_model_tree* parent,
    FILE*                           fp,
    const char*                     model_dirname
)
{
    uint32_t s, num_subtrees;
    uint32_t h, num_haplos; 
    uint32_t parent_haplo;
    char     haplo_buf[256] = {0};
    char     model_gmm_fname[1024] = {0};
    char     model_bmm_fname[1024] = {0};

    Gauss_berno_mixture_model_tree*  subtree;
    Gauss_berno_mixture_model_tree** subtrees;
    Gauss_berno_mixture_model*       model  = NULL;
    Vector_u32*                haplos = NULL;
    String*                    str = NULL;
    Error*                     e;


    if (*subtree_out != NULL)
    {
        free_gauss_berno_mixture_model_tree(*subtree_out);
        *subtree_out = NULL;
    }

    assert(*subtree_out = malloc(sizeof(Gauss_berno_mixture_model_tree)));
    subtree = *subtree_out;


    if (!skip_fp_spaces_and_comments(fp))
    {
        return create_error(ERROR_IO, __FILE__,
                "read_gauss_berno_mixture_model_tree", __LINE__, 
                "Missing model subtree");
    }

    /* Read the parent haplo type. */
    if (fscanf(fp, "%255s", haplo_buf) != 1)
    {
        return create_error(ERROR_INV_ARG, __FILE__, 
                "read_gauss_berno_mixture_model_subtree", __LINE__, 
                "Missing parent haplogroup for subtree");
    }
    if ((e = lookup_haplo_group_index_from_label(&parent_haplo, 
                    (const char*)haplo_buf)) != NULL)
    {
        return e;
    }


    /* Read the number of haplos in the root node and allocate space for them.*/
    if (fscanf(fp, "%u", &num_haplos) != 1 || num_haplos == 0)
    {
        return create_error(ERROR_INV_ARG, __FILE__, 
                "read_gauss_berno_mixture_model_subtree", __LINE__, 
                "Invalid number of haplos in model subtree");
    }
    create_vector_u32(&haplos, num_haplos);


    /* Read the number of subtrees in the node and allocate space for them. */
    if (fscanf(fp, "%u", &num_subtrees) != 1)
    {
        return create_error(ERROR_INV_ARG, __FILE__, 
                "read_gauss_berno_mixture_model_subtree", __LINE__, 
                "Invalid number of subtrees in model subtree");
    }
    subtrees = malloc(num_subtrees*sizeof(struct Gauss_berno_mixture_model_tree*));
    assert(subtrees);


    /* Read each haplo name and the model name. */
    if (!skip_fp_spaces_and_comments(fp))
    {
        return create_error(ERROR_IO, __FILE__,
                "read_gauss_berno_mixture_model_subtree", __LINE__, 
                "Missing haplogroup in model subtree");
    }
    for(h = 0; h < num_haplos; h++)
    {
        if (fscanf(fp, "%255s", haplo_buf) != 1)
        {
            return create_error(ERROR_INV_ARG, __FILE__, 
                    "read_gauss_berno_mixture_model_subtree", __LINE__, 
                    "Missing haplogroup in model subtree");
        }
        if ((e = lookup_haplo_group_index_from_label(&(haplos->elts[ h ]), 
                        (const char*)haplo_buf)) != NULL)
        {
            return e;
        }
    }
    if (fscanf(fp, "%255s", haplo_buf) != 1)
    {
        return create_error(ERROR_INV_ARG, __FILE__, 
                "read_gauss_berno_mixture_model_subtree", __LINE__, 
                "Missing GMM model file name in model subtree");
    }
    if (model_dirname)
    {
        snprintf(model_gmm_fname, 1023, "%s/%s", model_dirname, haplo_buf);
    }
    else
    {
        snprintf(model_gmm_fname, 1023, "%s", haplo_buf);
    }
    if (fscanf(fp, "%255s", haplo_buf) != 1)
    {
        return create_error(ERROR_INV_ARG, __FILE__, 
                "read_gauss_berno_mixture_model_subtree", __LINE__, 
                "Missing BMM model file name in model subtree");
    }
    if (model_dirname)
    {
        snprintf(model_bmm_fname, 1023, "%s/%s", model_dirname, haplo_buf);
    }
    else
    {
        snprintf(model_bmm_fname, 1023, "%s", haplo_buf);
    }
    if ((e = read_gauss_berno_mixture_model(&model, model_gmm_fname, 
                    model_bmm_fname)) != NULL)
    {
        string_printf(&str, 256, "%s: %s", model_gmm_fname, e->msg);
        return create_error(ERROR_INV_ARG, __FILE__, 
                "read_gauss_berno_mixture_model_subtree", __LINE__, str->chars);
    }


    subtree->parent_haplo = parent_haplo;
    subtree->parent       = parent;
    subtree->num_subtrees = num_subtrees;
    subtree->subtrees     = subtrees;
    subtree->haplos       = haplos;
    subtree->model        = model;

    /* Read the subtrees. */
    for(s = 0; s < num_subtrees; s++)
    {
        subtrees[ s ] = NULL;
        if ((e = read_gauss_berno_mixture_model_subtree(&(subtrees[ s ]),
                        subtree, fp, model_dirname)) != NULL)
        {
            return e;
        }
    }

    return NULL;
}


/**
 * @param  tree_out       Result parameter. If @em *tree_out is NULL, a model 
 *                        tree is allocated; otherwise it is first freed.
 * @param  tree_fname     File to read the model tree from.
 * @param  model_dirname  Directory to read the models from. If NULL, the
 *                        current working directory is used.
 *
 * @return On success, NULL is returned; otherwise an error is returned, but
 * the result parameters are not freed.
 */
Error* read_gauss_berno_mixture_model_tree
(
    Gauss_berno_mixture_model_tree** tree_out,
    const char*                tree_fname,
    const char*                model_dirname
)
{
    uint32_t s, num_subtrees;
    uint32_t h, num_haplos; 
    FILE*    fp;
    char     c;
    char     haplo_buf[256] = {0};
    char     model_gmm_fname[1024] = {0};
    char     model_bmm_fname[1024] = {0};

    Gauss_berno_mixture_model_tree*  tree;
    Gauss_berno_mixture_model_tree** subtrees;
    Gauss_berno_mixture_model*       model  = NULL;
    Vector_u32*                haplos = NULL;
    String*                    str = NULL;
    Error*                     e;


    if (*tree_out != NULL)
    {
        free_gauss_berno_mixture_model_tree(*tree_out);
        *tree_out = NULL;
    }

    if ((fp = fopen(tree_fname, "r")) == NULL)
    {
        return create_error(ERROR_IO, __FILE__, "read_gauss_berno_mixture_model_tree",
                __LINE__, strerror(errno));
    }

    assert(*tree_out = malloc(sizeof(Gauss_berno_mixture_model_tree)));
    tree = *tree_out;


    if (!skip_fp_spaces_and_comments(fp))
    {
        return create_error(ERROR_IO, __FILE__, "read_gauss_berno_mixture_model_tree",
                __LINE__, "Nothing to read in model tree");
    }

    if ((c = fgetc(fp)) != '*')
    {
        return create_error(ERROR_IO, __FILE__, "read_gauss_berno_mixture_model_tree",
                __LINE__, "Missing root in model tree");
    }


    /* Read the number of haplos in the root node and allocate space for them.*/
    if (fscanf(fp, "%u", &num_haplos) != 1 || num_haplos == 0)
    {
        return create_error(ERROR_INV_ARG, __FILE__,
                "read_gauss_berno_mixture_model_tree", __LINE__, 
                "Invalid number of haplos in model tree root");
    }
    create_vector_u32(&haplos, num_haplos);


    /* Read the number of subtrees in the root node and allocate space for
     * them. */
    if (fscanf(fp, "%u", &num_subtrees) != 1)
    {
        return create_error(ERROR_INV_ARG, __FILE__, 
                "read_gauss_berno_mixture_model_tree", __LINE__, 
                "Invalid number of subtrees in model tree root");
    }
    subtrees = malloc(num_subtrees*sizeof(struct Gauss_berno_mixture_model_tree*));
    assert(subtrees);


    /* Read each haplo name and the model name. */
    if (!skip_fp_spaces_and_comments(fp))
    {
        return create_error(ERROR_INV_ARG, __FILE__, 
                "read_gauss_berno_mixture_model_tree", __LINE__, 
                "Missing haplogroup in model tree");
    }
    for(h = 0; h < num_haplos; h++)
    {
        if (fscanf(fp, "%255s", haplo_buf) != 1)
        {
            return create_error(ERROR_INV_ARG, __FILE__, 
                    "read_gauss_berno_mixture_model_tree", __LINE__, 
                    "Missing haplogroup in model tree root");
        }
        if ((e = lookup_haplo_group_index_from_label(&(haplos->elts[ h ]), 
                        (const char*)haplo_buf)) != NULL)
        {
            return e;
        }
    }
    if (fscanf(fp, "%255s", haplo_buf) != 1)
    {
        return create_error(ERROR_INV_ARG, __FILE__, 
                "read_gauss_berno_mixture_model_tree", __LINE__, 
                "Missing GMM model file name in model tree root");
    }
    if (model_dirname)
    {
        snprintf(model_gmm_fname, 1023, "%s/%s", model_dirname, haplo_buf);
    }
    else
    {
        snprintf(model_gmm_fname, 1023, "%s", haplo_buf);
    }
    if (fscanf(fp, "%255s", haplo_buf) != 1)
    {
        return create_error(ERROR_INV_ARG, __FILE__, 
                "read_gauss_berno_mixture_model_tree", __LINE__, 
                "Missing BMM model file name in model tree root");
    }
    if (model_dirname)
    {
        snprintf(model_bmm_fname, 1023, "%s/%s", model_dirname, haplo_buf);
    }
    else
    {
        snprintf(model_bmm_fname, 1023, "%s", haplo_buf);
    }
    if ((e = read_gauss_berno_mixture_model(&model, model_gmm_fname, 
                    model_bmm_fname)) != NULL)
    {
        string_printf(&str, 256, "%s: %s", model_gmm_fname, e->msg);
        return create_error(ERROR_INV_ARG, __FILE__,
                "read_gauss_berno_mixture_model_tree", __LINE__, str->chars); 
    }


    tree->parent_haplo = 0;
    tree->parent       = NULL;
    tree->num_subtrees = num_subtrees;
    tree->subtrees     = subtrees;
    tree->haplos       = haplos;
    tree->model        = model;

    /* Read the subtrees. */
    for(s = 0; s < num_subtrees; s++)
    {
        subtrees[ s ] = NULL;
        if ((e = read_gauss_berno_mixture_model_subtree(&(subtrees[ s ]),
                        tree, fp, model_dirname)) != NULL)
        {
            return e;
        }
    }

    if (fclose(fp) != 0)
    {
        return create_error(ERROR_IO, __FILE__,
                "read_gauss_berno_mixture_model_tree", __LINE__,
                strerror(errno));
    }

    return NULL;
}





/** @param  tree  Model tree to free. */
void free_gauss_berno_mixture_model_tree(Gauss_berno_mixture_model_tree* tree)
{
    uint32_t s;

    if (!tree)
        return;

    for(s = 0; s < tree->num_subtrees; s++)
    {
        free_gauss_berno_mixture_model_tree(tree->subtrees[ s ]);
    }

    free(tree->subtrees);
    free_vector_u32(tree->haplos);
    free_gauss_berno_mixture_model(tree->model);
    free(tree);
}




/** 
 * @brief Tests a predicted haplogroup against an actual haplogroup name based
 * on substring prefixes.
 */
static Gauss_berno_mixture_accuracy test_haplogroup_substrings
(
    uint32_t pred_haplo,
    uint32_t actual_haplo
)
{
    const char* pred_haplo_name;
    const char* actual_haplo_name;

    assert(lookup_haplo_group_label_from_index(&pred_haplo_name, pred_haplo) 
            == NULL);
    assert(lookup_haplo_group_label_from_index(&actual_haplo_name, actual_haplo)
            == NULL);

    if (strcmp(pred_haplo_name, actual_haplo_name) == 0)
    {
        return GAUSS_BERNO_MIX_CORRECT;
    }
    if (strncmp(pred_haplo_name, actual_haplo_name, strlen(pred_haplo_name)) 
            == 0)
    {
        return GAUSS_BERNO_MIX_UP_TREE;
    }
    if (strncmp(pred_haplo_name, actual_haplo_name, strlen(actual_haplo_name)) 
            == 0)
    {
        return GAUSS_BERNO_MIX_DOWN_TREE;
    }

    return GAUSS_BERNO_MIX_INCORRECT;
}


/** @brief Recursively searches a model tree for a haplogroup. */
static const Gauss_berno_mixture_model_tree* recursively_find_haplo
(
    uint32_t                        haplo,
    const Gauss_berno_mixture_model_tree* tree
)
{
    uint32_t h, num_haplos;
    uint32_t s, num_subtrees;
    Gauss_berno_mixture_accuracy accuracy;
    const Gauss_berno_mixture_model_tree* subtree;

    num_haplos   = tree->haplos->num_elts;
    num_subtrees = tree->num_subtrees;

    for (h = 0; h < num_haplos; h++)
    {
        accuracy = test_haplogroup_substrings(haplo, tree->haplos->elts[ h ]);
        if (accuracy == GAUSS_BERNO_MIX_CORRECT || 
                accuracy == GAUSS_BERNO_MIX_UP_TREE)
        {
            return tree;
        }
        if (accuracy == GAUSS_BERNO_MIX_DOWN_TREE)
        {
            for (s = 0; s < num_subtrees; s++)
            {
                if (tree->subtrees[s]->parent_haplo == tree->haplos->elts[h])
                {
                    if ((subtree = recursively_find_haplo(haplo, 
                                tree->subtrees[ s ])) != NULL)
                    {
                        return subtree;
                    }
                }
            }
            return tree;
        }
    }

    for (s = 0; s < num_subtrees; s++)
    {
        if ((subtree = recursively_find_haplo(haplo, tree->subtrees[ s ])) 
                != NULL)
        {
            return subtree;
        }
    }

    return NULL;
}


/** 
 * @brief Returns whether or not @em parent is the parent tree of 
 * @em subtree.
 */
static uint8_t is_parent_of_subtree
(
    uint32_t                        parent_haplo,
    const Gauss_berno_mixture_model_tree* parent,
    const Gauss_berno_mixture_model_tree* subtree
)
{
    uint32_t num_haplos;
    uint32_t h;
    uint32_t haplo;

    if (parent == subtree)
    {
        return 0;
    }

    num_haplos = parent->haplos->num_elts;

    for (h = 0; h < num_haplos; h++)
    {
        haplo = parent->haplos->elts[ h ];
        if (subtree->parent_haplo == haplo &&
            test_haplogroup_substrings(parent_haplo, haplo) 
                 != GAUSS_BERNO_MIX_INCORRECT)
        {
            return 1;
        }
    }
    if (subtree->parent != NULL)
    {
        return is_parent_of_subtree(parent_haplo, parent, subtree->parent);
    }
    return 0;
}


/** 
 * @param  pred_haplo    Predicted haplogroup.
 * @param  actual_haplo  Actual haplogroup.
 * @param  tree          Model tree.
 *
 * @return Returns the accuracy of the prediction.
 */
Gauss_berno_mixture_accuracy gauss_berno_mixture_model_tree_haplo_accuracy
(
    uint32_t                        pred_haplo,
    uint32_t                        actual_haplo,
    const Gauss_berno_mixture_model_tree* tree
)
{
    const Gauss_berno_mixture_model_tree* pred_subtree;
    const Gauss_berno_mixture_model_tree* actual_subtree;
    Gauss_berno_mixture_accuracy result;

    if ((result = test_haplogroup_substrings(pred_haplo, actual_haplo)) 
            != GAUSS_BERNO_MIX_INCORRECT)
    {
        return result;
    }

    assert(pred_subtree = recursively_find_haplo(pred_haplo, tree));
    assert(actual_subtree = recursively_find_haplo(actual_haplo, tree));

    if (is_parent_of_subtree(pred_haplo, pred_subtree, actual_subtree))
    {
        result = GAUSS_BERNO_MIX_UP_TREE;
    }
    else if (is_parent_of_subtree(actual_haplo, actual_subtree, pred_subtree))
    {
        result = GAUSS_BERNO_MIX_DOWN_TREE;
    }

    return result;
}
