/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Gaussian and Bernoulli mixture model classifier.
 */


#ifndef HAPLO_GMM_BMM_H
#define HAPLO_GMM_BMM_H


#include <config.h>

#include <stdlib.h>
#include <inttypes.h>

#include <jwsc/base/error.h>
#include <jwsc/vector/vector.h>
#include <jwsc/matrix/matrix.h>
#include <jwsc/matblock/matblock.h>

#include "gmm.h"
#include "bmm.h"


/** 
 * @brief Gaussian and Bernoulli mixture model.
 *
 * A GMM and BMM is created conditioned on each haplogroup; the GMM and BMM
 * parameters (mu, sigma and pi) are actually in arrays indexed by haplogroup.
 */
typedef struct
{
    /** @brief Gaussian mixture model. */
    Gauss_mixture_model* gmm;

    /** @breif Bernoulli mixture model. */
    Berno_mixture_model* bmm;
} 
Gauss_berno_mixture_model;


/** @brief Gaussian and Bernoulli mixture model tree. */
typedef struct Gauss_berno_mixture_model_tree
{
    /** @brief Parent haplogroup name. */
    uint32_t parent_haplo;

    /** @brief Pointer to the parent haplogroup model tree. */
    const struct Gauss_berno_mixture_model_tree*  parent;

    /** @brief Number of model subtrees. */
    uint32_t  num_subtrees;

    /** @brief Array of model subtrees pointers. */
    struct Gauss_berno_mixture_model_tree** subtrees;

    /** @brief Haplogroups in this model tree node. */
    Vector_u32* haplos;

    /** @brief Gaussian mixture model. */
    Gauss_berno_mixture_model* model;
} 
Gauss_berno_mixture_model_tree;


void split_gmm_bmm_markers
(
    Matrix_i32**      gmm_markers_out,
    Matrix_i32**      bmm_markers_out,
    const Matrix_i32* markers,
    uint32_t          num_gmm_markers
);


/** @brief Trains a Gaussian and Bernoulli mixture model. */
void train_gauss_berno_mixture_model
(
    Gauss_berno_mixture_model** model_out,
    const Vector_u32*     haplos, 
    const Matrix_i32*     markers,
    uint32_t              num_gmm_markers,
    uint32_t              num_gmm_mix,
    uint32_t              num_bmm_mix,
    const Vector_d*       haplo_priors
);


/** 
 * @brief Trains a set of Gaussian and Bernoulli mixture models in a model
 * information file. 
 */
Error* train_gauss_berno_mixture_model_info
(
    const Vector_u32* haplos, 
    const Matrix_i32* markers,
    uint32_t          num_gmm_markers,
    const char*       info_fname,
    const char*       model_dirname
);


/** @brief Writes a Gaussian and Bernoulli mixture model. */
Error* write_gauss_berno_mixture_model
(
    Gauss_berno_mixture_model* model, 
    const char* gmm_fname,
    const char* bmm_fname
);


/** @brief Reads a Gaussian and Bernoulli mixture model. */
Error* read_gauss_berno_mixture_model
(
    Gauss_berno_mixture_model** model_out, 
    const char* gmm_fname,
    const char* bmm_fname
);


/** @brief Frees a Gaussian and Bernoulli mixture model. */
void free_gauss_berno_mixture_model(Gauss_berno_mixture_model* model);


/** 
 * @brief Predicts a haplogroup for a marker sample using a Gaussian and
 * Bernoulli mixture model.
 */
Error* predict_haplo_with_gauss_berno_mixture_model
(
    uint32_t*                  haplo_out,
    double*                    confidence_out,
    const Vector_i32*          gmm_markers,
    const Vector_i32*          bmm_markers,
    const Gauss_berno_mixture_model* model
);


/** 
 * @brief Predicts the haplogroups for a set of marker samples using a
 * Gaussian and bernoulli mixture model.
 */
Error* predict_haplos_with_gauss_berno_mixture_model
(
    Vector_u32**               haplos_out,
    Vector_d**                 confidence_out,
    const Matrix_i32*          markers,
    uint32_t                   num_gmm_markers,
    const Gauss_berno_mixture_model* model
);


/** 
 * @brief Predicts the haplogroups for a set of marker samples using a 
 * Gaussian and Bernoulli mixture model tree.
 */
Error* predict_haplos_with_gauss_berno_mixture_model_tree
(
    Vector_u32**                    haplos_out,
    Vector_d**                      confidence_out,
    const Matrix_i32*               markers,
    uint32_t                        num_gmm_markers,
    const Gauss_berno_mixture_model_tree* tree
);


/** @brief Reads a Gaussian and Bernoulli mixture model tree from. */
Error* read_gauss_berno_mixture_model_tree
(
    Gauss_berno_mixture_model_tree** tree_out,
    const char*                tree_fname,
    const char*                model_dirname
);


/** @brief Frees a Gaussian and Bernoulli mixture model tree. */
void free_gauss_berno_mixture_model_tree(Gauss_berno_mixture_model_tree* tree);


/** @brief Haplogroup prediction accuracy for a Gaussian mixture model. */
typedef enum 
{
    GAUSS_BERNO_MIX_INCORRECT,
    GAUSS_BERNO_MIX_CORRECT,
    GAUSS_BERNO_MIX_UP_TREE,
    GAUSS_BERNO_MIX_DOWN_TREE
}
Gauss_berno_mixture_accuracy;


/** 
 * @brief Tests the accuracy of a predicted haplogroup under a Gaussian
 * and Bernoulli mixture model tree.
 */
Gauss_berno_mixture_accuracy gauss_berno_mixture_model_tree_haplo_accuracy
(
    uint32_t                        pred_haplo,
    uint32_t                        actual_haplo,
    const Gauss_berno_mixture_model_tree* tree
);


#endif
