/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Impute missing marker values in a set of Y-STR samples.
 *
 * A missing marker value is defined as < 1.
 *
 * Available imputation types include averaging if the sample labels are
 * known or interpolating values from nearest neighbors.
 */


#include <config.h>

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <inttypes.h>
#include <string.h>

#include <libxml/tree.h>

#ifdef HAPLO_HAVE_DMALLOC
#include <dmalloc.h>
#endif

#include <jwsc/base/error.h>
#include <jwsc/base/option.h>
#include <jwsc/vector/vector.h>
#include <jwsc/matrix/matrix.h>
#include <jwsc/matrix/matrix_io.h>
#include <jwsc/matblock/matblock.h>
#include <jwsc/stat/kmeans.h>
#include <jwsc/stat/gmm.h>

#include "haplo_groups.h"
#include "options.h"
#include "output.h"
#include "input.h"
#include "xml.h"


#define  NUM_OPTS_NO_ARG    0 + NUM_SHARED_OPTS_NO_ARG
#define  NUM_OPTS_WITH_ARG  8 + NUM_SHARED_OPTS_WITH_ARG


#define  LABEL_COL     0
#define  IMPUTE_TYPE   HAPLO_IMPUTE_AVG
#define  OUTPUT_FNAME  "/dev/stdout"


/** @brief Types of clustering supported. */
typedef enum
{
    HAPLO_IMPUTE_AVG,
    HAPLO_IMPUTE_NN
}
Haplo_impute_type;


/** @brief Program options not taking an argument. */
Option_no_arg opts_no_arg[NUM_OPTS_NO_ARG];

/** @brief Program options taking an argument. */
Option_with_arg opts_with_arg[NUM_OPTS_WITH_ARG];

/** @brief Type of imputation to use. */
static Haplo_impute_type impute_type = IMPUTE_TYPE;

/** @brief Output file name. */
static const char* output_fname = OUTPUT_FNAME;


/** @brief Returns the number of options not taking an argument. */
uint32_t get_num_opts_no_arg()
{
    return NUM_OPTS_NO_ARG;
}

/** @brief Returns the number of options taking an argument. */
uint32_t get_num_opts_with_arg()
{
    return NUM_OPTS_WITH_ARG;
}

/** @brief Prints the program usage to stderr. */
void print_usage()
{
    fprintf(stderr, "usage: haplo-impute OPTIONS [data-fname | <stdin>]\n");
    print_options(stderr, 27, NUM_OPTS_NO_ARG, opts_no_arg, NUM_OPTS_WITH_ARG,
            opts_with_arg);
}

/** @brief Process the 'impute-type' program option. */
static Error* process_impute_type_opt(Option_arg arg)
{
    if (arg == NULL)
    {
        return JWSC_EARG("Option 'impute-type' requires an argument");
    }
    if (strncmp(arg, "avg", 3) == 0)
    {
        impute_type = HAPLO_IMPUTE_AVG;
    }
    else if (strncmp(arg, "nn", 2) == 0)
    {
        impute_type = HAPLO_IMPUTE_NN;
    }
    else
    {
        return JWSC_EARG("Option 'impute-type' must be one of {avg, nn}");
    }
    return NULL;
}

/** @brief Process the 'output' program option. */
Error* process_output_opt(Option_arg arg)
{
    if (arg == NULL)
    {
        return JWSC_EARG("Option 'output' requires an argument");
    }
    output_fname = arg;
    return NULL;
}

/** @brief Initializes the haplo-impute program options. */
static void init_impute_options()
{
    uint32_t i;

    char s_name;
    const char* l_name;
    const char* desc;

    Error* (*farg)(const char*);

    init_options(opts_no_arg, opts_with_arg);

    opts.label_col = LABEL_COL;

    i = NUM_SHARED_OPTS_WITH_ARG;
    l_name = "output";
    s_name = 0;
    desc   = "File to write the imputed output to. Default is stdout.";
    farg   = process_output_opt;
    init_option_with_arg(&(opts_with_arg[i++]), l_name, s_name, desc, farg);

    l_name = "aux-input";
    s_name = 0;
    desc   = "Auxiliary input file. Provides marker information used for imputing missing values.";
    farg   = process_aux_input_opt;
    init_option_with_arg(&(opts_with_arg[i++]), l_name, s_name, desc, farg);

    l_name = "aux-id-cols";
    s_name = 0;
    desc   = "Auxiliary input id-cols. Comma separated ordered list of columns to use for sample identification. Prefixes the output of each sample. Count begins with 1 at the first column of the file. Set to zero to ignore the id column";
    farg   = process_aux_id_cols_opt;
    init_option_with_arg(&(opts_with_arg[i++]), l_name, s_name, desc, farg);

    l_name = "aux-label-col";
    s_name = 0;
    desc   = "Auxiliary input label-col. Column containing the haplo group labels. Count begins with 1 at the first column of the file. Set to zero to ignore the label column.";
    farg   = process_aux_label_col_opt;
    init_option_with_arg(&(opts_with_arg[i++]), l_name, s_name, desc, farg);

    l_name = "aux-1st-marker-col";
    s_name = 0;
    desc   = "Auxiliary input 1st-marker-col. Column containing the first marker. Use in conjunction with num-markers to specify the markers for reading. All other markers are assumed to follow this one. Count begins with 1.";
    farg   = process_aux_first_marker_col_opt;
    init_option_with_arg(&(opts_with_arg[i++]), l_name, s_name, desc, farg);

    l_name = "aux-num-markers";
    s_name = 0;
    desc   = "Auxiliary input num-markers. Number of markers to read. Use in conjunction with 1st-marker-col to specify the markers for reading.";
    farg   = process_aux_num_markers_opt;
    init_option_with_arg(&(opts_with_arg[i++]), l_name, s_name, desc, farg);

    l_name = "aux-marker-cols";
    s_name = 0;
    desc   = "Auxiliary input marker-cols. Comma separated ordered list of markers to use for training. Use instead of 1st-marker-col and num-markers. Count begins with 1 at the first column of the CSV file.";
    farg   = process_aux_marker_cols_opt;
    init_option_with_arg(&(opts_with_arg[i++]), l_name, s_name, desc, farg);

    l_name = "impute-type";
    s_name = 0;
    desc   = "Type of imputation to use. Must be one of {avg, nn}.";
    farg   = process_impute_type_opt;
    init_option_with_arg(&(opts_with_arg[i++]), l_name, s_name, desc, farg);
    assert(i == NUM_OPTS_WITH_ARG);
}

/** @brief Write the output file header. */
static void write_header
(
    const Matblock_u8* ids,
    const Vector_u32*  labels,
    const Matrix_i32*  markers,
    FILE*              fp
)
{
    uint32_t i, j;

    if (!opts.header_out || opts.output_format == HAPLO_OUTPUT_XML)
        return;

    if (ids)
    {
        for (i = 0; i < ids->num_rows; i++)
        {
            switch (opts.output_format)
            {
                case HAPLO_OUTPUT_TXT:
                    fprintf(fp, "ID %-7d  ", i+1);
                    break;
                case HAPLO_OUTPUT_CSV:
                    fprintf(fp, "ID %d,", i+1);
                    break;
                case HAPLO_OUTPUT_XML:
                    break;
            }
        }
    }

    if (labels)
    {
        switch (opts.output_format)
        {
            case HAPLO_OUTPUT_TXT:
                fprintf(fp, "%-10s  ", "Label");
                break;
            case HAPLO_OUTPUT_CSV:
                fprintf(fp, "%s,", "Label");
                break;
            case HAPLO_OUTPUT_XML:
                break;
        }
    }

    switch (opts.output_format)
    {
        case HAPLO_OUTPUT_TXT:
            fprintf(fp, "%6s %2d", "Marker", 1);
            for (j = 1; j < markers->num_cols; j++)
            {
                fprintf(fp, "  %6s %2d", "Marker", j+1);
            }
            break;
        case HAPLO_OUTPUT_CSV:
            fprintf(fp, "%s %d", "Marker", 1);
            for (j = 1; j < markers->num_cols; j++)
            {
                fprintf(fp, ",%s %d", "Marker", j+1);
            }
            break;
        case HAPLO_OUTPUT_XML:
            break;
    }
    fprintf(fp, "\n");
}

/** @brief Writes the output imputed data. */
static void write_imputed_data
(
    const Matblock_u8* ids,
    const Vector_u32*  labels,
    const Matrix_i32*  markers
)
{
    FILE*       fp;
    uint32_t    i, j;
    xmlDoc*     xml_doc   = NULL;
    xmlNode*    xml_root  = NULL;
    xmlNode*    xml_node  = NULL;
    xmlNode*    xml_child = NULL;
    char*       xml_buf   = NULL;
    uint32_t    xml_len;
    Error*      err;

    if ((err = open_output(&fp, &xml_doc, "haplo-input", "haplo-input.dtd", 
                    output_fname)))
    {
        print_error_msg("haplo-impute", err->msg);
    }

    write_header(ids, labels, markers, fp);

    if (opts.output_format == HAPLO_OUTPUT_XML)
    {
        xml_root = xmlDocGetRootElement(xml_doc);
    }

    for (i = 0; i < markers->num_rows; i++)
    {
        if (opts.output_format == HAPLO_OUTPUT_XML)
        {
            xml_node = XMLNewChild(xml_root, "sample", NULL);
            xml_len = 256;
            xml_buf = malloc(xml_len*sizeof(char));
            snprintf(xml_buf, xml_len, "%d", i+1);
            XMLNewProp(xml_node, "number", xml_buf);
            free(xml_buf);
        }

        write_ids(ids, i, HAPLO_SEP_SUFFIX, fp, xml_node);
        write_label(labels, i, HAPLO_SEP_SUFFIX, fp, xml_node);

        switch (opts.output_format)
        {
            case HAPLO_OUTPUT_TXT:
                fprintf(fp, "%-9d", markers->elts[ i ][ 0 ]);
                for (j = 1; j < markers->num_cols; j++)
                {
                    fprintf(fp, "  %-9d", markers->elts[ i ][ j ]);
                }
                fprintf(fp, "\n");
                break;
            case HAPLO_OUTPUT_CSV:
                fprintf(fp, "%d", markers->elts[ i ][ 0 ]);
                for (j = 1; j < markers->num_cols; j++)
                {
                    fprintf(fp, ",%d", markers->elts[ i ][ j ]);
                }
                fprintf(fp, "\n");
                break;
            case HAPLO_OUTPUT_XML:
                xml_len = 256;
                xml_buf = malloc(xml_len*sizeof(char));
                for (j = 0; j < markers->num_cols; j++)
                {
                    snprintf(xml_buf, xml_len, "%d", markers->elts[i][j]);
                    xml_child = XMLNewChild(xml_node, "marker", xml_buf);
                    snprintf(xml_buf, xml_len, "%d", j+1);
                    XMLNewProp(xml_child, "number", xml_buf);
                }
                free(xml_buf);
                break;
        }
    }

    if ((err = close_output(fp, xml_doc, output_fname)))
    {
        print_error_msg("haplo-impute", err->msg);
    }
}


/** @brief Main function for the @em haplo-impute program. */
int main(int argc, const char** argv)
{
    int         argi;
    const char* data_fname = "/dev/stdin";
    Error*      err;

    Matblock_u8* imp_ids       = NULL;
    Vector_u32*  imp_labels    = NULL;
    Matrix_i32*  imp_markers   = NULL;
    Matblock_u8* src_ids       = NULL;
    Vector_u32*  src_labels    = NULL;
    Matrix_i32*  src_markers   = NULL;

    init_impute_options();

    if ((err = process_options(argc, argv, &argi, NUM_OPTS_NO_ARG, opts_no_arg,
                    NUM_OPTS_WITH_ARG, opts_with_arg)) != NULL)
    {
        print_error_msg_exit("haplo-impute", err->msg);
    }

    if ((argc - argi) == 1)
    {
        data_fname = argv[ argi ];
    }

    if ((err = read_haplo_groups(opts.labels_fname)))
    {
        print_error_msg_exit("haplo-impute", err->msg);
    }

    if ((err = read_input(&imp_ids, &imp_labels, &imp_markers, data_fname)))
    {
        print_error_msg_exit("haplo-impute", err->msg);
    }

    if (opts.aux_input_fname)
    {
        if ((err = read_aux_input(&src_ids, &src_labels, &src_markers,
                        opts.aux_input_fname)))
        {
            print_error_msg_exit("haplo-impute", err->msg);
        }
    }

    switch (impute_type)
    {
        case HAPLO_IMPUTE_AVG:
            if (imp_labels && src_labels)
            {
                if ((err = impute_missing_markers_from_avg(imp_labels,
                                imp_markers, src_labels, src_markers)))
                {
                    print_error_msg_exit("haplo-impute", err->msg);
                }
            }
            else if (imp_labels)
            {
                if ((err = impute_missing_markers_from_avg(imp_labels,
                                imp_markers, imp_labels, imp_markers)))
                {
                    print_error_msg_exit("haplo-impute", err->msg);
                }
            }
            else
            {
                print_error_msg_exit("haplo-impute", 
                        "No labels for avg imputing");
            }
            break;
        case HAPLO_IMPUTE_NN:
            if (src_markers)
            {
                if ((err = impute_missing_markers_from_nn(imp_markers, 
                                src_markers)))
                {
                    print_error_msg_exit("haplo-impute", err->msg);
                }
            }
            else
            {
                if ((err = impute_missing_markers_from_nn(imp_markers, 
                                imp_markers)))
                {
                    print_error_msg_exit("haplo-impute", err->msg);
                }
            }
            break;
    }

    write_imputed_data(imp_ids, imp_labels, imp_markers);

    free_matblock_u8(imp_ids);
    free_vector_u32(imp_labels);
    free_matrix_i32(imp_markers);
    free_matblock_u8(src_ids);
    free_vector_u32(src_labels);
    free_matrix_i32(src_markers);

    if (get_num_unhandled_errors() > 0)
    {
        print_error_msg_exit("haplo-impute", "Unhandled errors");
    }

    return EXIT_SUCCESS;
}
