/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @mainpage
 *
 * Predicts haplogroup using models trained with Y-STR data.
 *
 * Licensed under <a href='http://creativecommons.org/licenses/by-nc-sa/3.0/us/'>Creative Commons BY-NC-SA 3.0</a>.
 *
 * Questions or comments? Contact <a href='http://vision.cs.arizona.edu/schlecht'>Joseph Schlecht</a>.
 *
 * <hr>
 *
 * Use @c 'haplo-train' to train the models.
 * @verbatim
usage: haplo-train OPTIONS [data-fname | <stdin>]
 -h, --help                Prints program usage.
 -v, --version             Prints program version.
     --header-in           The input data contains a header (descriptive) line,
                           which should be discarded.
     --header-out          Write a header (descriptive) line to the first line
                           of the output results.
     --options=ARG         File containing program options. Any options
                           appearing on the command line following this option
                           take precendence over those in the options file.
     --seed=ARG            Random seed.
     --input-format=ARG    Input file format. Must be one of {txt, csv, xml}. If
                           the input is XML, it must conform to the XML DTD
                           haplo-input.dtd.
     --input-dtd=ARG       If the input format is XML, validate it with this
                           DTD.
     --output-format=ARG   Output file format. Must be one of {txt, csv, xml}.
     --labels=ARG          XML file containing the organization and listing of
                           possible haplo groups labels for the samples. Must
                           conform to the XML DTD haplo-labels.dtd.
     --labels-dtd=ARG      Validate the XML labels file with this DTD.
     --id-cols=ARG         Comma separated ordered list of columns to use for
                           sample identification. Prefixes the output of each
                           sample. Count begins with 1 at the first column of
                           the file. Set to zero to ignore the id column
     --label-col=ARG       Column containing the haplo group labels. Count
                           begins with 1 at the first column of the file. Set
                           to zero to ignore the label column.
     --1st-marker-col=ARG  Column containing the first marker. Use in
                           conjunction with num-markers to specify the markers
                           for reading. All other markers are assumed to follow
                           this one. Count begins with 1.
     --num-markers=ARG     Number of markers to read. Use in conjunction with
                           1st-marker-col to specify the markers for reading.
     --marker-cols=ARG     Comma separated ordered list of markers to use for
                           training. Use instead of 1st-marker-col and
                           num-markers. Count begins with 1 at the first column
                           of the CSV file.
     --model-dir=ARG       Directory to put trained models in.
     --data-out-dir=ARG    Directory to put the generated training data in for
                           each model. The name of the model is used, so if
                           this directory is set as the same as the model-dir,
                           the models could be overwritten. The default is not
                           to output the data.
     --nb-freq=ARG         Naive Bayes non-parametric frequency model tree
                           information.
     --nb-freq-dtd=ARG     Validate the naive Bayes non-parametric frequency
                           model tree information XML file with this DTD.
     --nb-gauss=ARG        Naive Bayes Gaussian model tree information.
     --nb-gauss-dtd=ARG    Validate the naive Bayes Gaussian model tree
                           information XML file with this DTD.
     --nb-gmm=ARG          Naive Bayes Gaussian mixture model tree information.
     --nb-gmm-dtd=ARG      Validate the naive Bayes Gaussian mixture model tree
                           information XML file with this DTD.
     --mv-gmm=ARG          Multivariate Gaussian mixture model tree information.
     --mv-gmm-dtd=ARG      Validate the multivariate Gaussian mixture model tree
                           information XML file with this DTD.
     --mv-mmm=ARG          Multivariate multinomial mixture model tree
                           information.
     --mv-mmm-dtd=ARG      Validate the multivariate multinomial mixture model
                           tree information XML file with this DTD.
     --svm=ARG             SVM model tree information.
     --svm-dtd=ARG         Validate the SVM model tree information XML file with
                           this DTD.
     --weka-j48=ARG        Weka J48 model tree information.
     --weka-part=ARG       Weka PART model tree information.
     --weka-jar=ARG        Weka java archive file. Required for using the Weka
                           algorithms.
     --weka-dtd=ARG        Validate the Weka model tree information XML files
                           with this DTD.
     --nearest=ARG         Nearest neighbor model information.
     --nearest-dtd=ARG     Validate the nearest neighbor model information XML
                           file with this DTD. @endverbatim
 *
 * <br>
 *
 * Use @c 'haplo-predict' to predict haplogroup with the trained models.
 * @verbatim
usage: haplo-predict OPTIONS [data-fname | <stdin>]
 -h, --help                Prints program usage.
 -v, --version             Prints program version.
     --header-in           The input data contains a header (descriptive) line,
                           which should be discarded.
     --header-out          Write a header (descriptive) line to the first line
                           of the output results.
     --exclude-one         When performing the tandem prediction decision,
                           exclude at most one prediction from the set of
                           classification algorithms. There must be three or
                           more algorithms in play for this to take effect.
     --options=ARG         File containing program options. Any options
                           appearing on the command line following this option
                           take precendence over those in the options file.
     --seed=ARG            Random seed.
     --input-format=ARG    Input file format. Must be one of {txt, csv, xml}. If
                           the input is XML, it must conform to the XML DTD
                           haplo-input.dtd.
     --input-dtd=ARG       If the input format is XML, validate it with this
                           DTD.
     --output-format=ARG   Output file format. Must be one of {txt, csv, xml}.
     --labels=ARG          XML file containing the organization and listing of
                           possible haplo groups labels for the samples. Must
                           conform to the XML DTD haplo-labels.dtd.
     --labels-dtd=ARG      Validate the XML labels file with this DTD.
     --id-cols=ARG         Comma separated ordered list of columns to use for
                           sample identification. Prefixes the output of each
                           sample. Count begins with 1 at the first column of
                           the file. Set to zero to ignore the id column
     --label-col=ARG       Column containing the haplo group labels. Count
                           begins with 1 at the first column of the file. Set
                           to zero to ignore the label column.
     --1st-marker-col=ARG  Column containing the first marker. Use in
                           conjunction with num-markers to specify the markers
                           for reading. All other markers are assumed to follow
                           this one. Count begins with 1.
     --num-markers=ARG     Number of markers to read. Use in conjunction with
                           1st-marker-col to specify the markers for reading.
     --marker-cols=ARG     Comma separated ordered list of markers to use for
                           training. Use instead of 1st-marker-col and
                           num-markers. Count begins with 1 at the first column
                           of the CSV file.
     --output=ARG          File to output the predictions to. The default is
                           stdout.
     --model-dir=ARG       Directory containing the trained models.
     --nb-freq=ARG         Naive Bayes non-parametric frequency model tree
                           information.
     --nb-freq-dtd=ARG     Validate the naive Bayes non-parametric frequency
                           model tree information XML file with this DTD.
     --nb-gauss=ARG        Naive Bayes Gaussian model tree information.
     --nb-gauss-dtd=ARG    Validate the naive Bayes Gaussian model tree
                           information XML file with this DTD.
     --nb-gmm=ARG          Naive Bayes Gaussian mixture model tree information.
     --nb-gmm-dtd=ARG      Validate the naive Bayes Gaussian mixture model tree
                           information XML file with this DTD.
     --mv-gmm=ARG          Multivariate Gaussian mixture model tree information.
     --mv-gmm-dtd=ARG      Validate the multivariate Gaussian mixture model tree
                           information XML file with this DTD.
     --svm=ARG             SVM model tree information.
     --svm-dtd=ARG         Validate the SVM model tree information XML file with
                           this DTD.
     --weka-j48=ARG        Weka J48 model tree information.
     --weka-part=ARG       Weka PART model tree information.
     --weka-jar=ARG        Weka java archive file. Required for using the Weka
                           algorithms.
     --weka-dtd=ARG        Validate the Weka model tree information XML files
                           with this DTD.
     --nearest=ARG         Nearest neighbor model information.
     --nearest-dtd=ARG     Validate the nearest neighbor model information XML
                           file with this DTD.
     --nearest-max-d=ARG   Maximum distance allowed for a nearest neighbor
                           classification. @endverbatim
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Trains a set of Y-STR haplotype label classifiers.
 *
 * Classifier types include
 * -# Naive Bayes non-parametric marker frequency model
 * -# Naive Bayes Gaussian model
 * -# Naive Bayes Gaussian mixture model
 * -# Multivariate Gaussian mixture model
 * -# Multivariate multinomial mixture model
 * -# Support Vector Machines
 * -# Weka PART and J48
 * -# Nearest neighbor
 */


#include <config.h>

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <errno.h>

#ifdef HAPLO_HAVE_DMALLOC
#include <dmalloc.h>
#endif

#include <jwsc/base/error.h>
#include <jwsc/base/option.h>
#include <jwsc/vector/vector.h>
#include <jwsc/matrix/matrix.h>
#include <jwsc/matblock/matblock.h>

#include "haplo_groups.h"
#include "options.h"
#include "input.h"
#include "nb_freq.h"
#include "nb_gauss.h"
#include "nb_gmm.h"
#include "mv_gmm.h"
#include "mv_mmm.h"
#ifdef HAPLO_ENABLE_SVM
#include "svm_tree.h"
#endif
#ifdef HAPLO_ENABLE_WEKA
#include "weka.h"
#endif
#include "nearest.h"


#ifdef HAPLO_ENABLE_SVM
#define NUM_SVM_OPTS 2
#else
#define NUM_SVM_OPTS 0
#endif

#ifdef HAPLO_ENABLE_WEKA
#define NUM_WEKA_OPTS 4
#else
#define NUM_WEKA_OPTS 0
#endif

#define  NUM_OPTS_NO_ARG    0 + NUM_SHARED_OPTS_NO_ARG
#define  NUM_OPTS_WITH_ARG  14 + NUM_SVM_OPTS + NUM_WEKA_OPTS + NUM_SHARED_OPTS_WITH_ARG


#define  LABEL_COL         1
#define  DATA_OUT_DIRNAME  0


/** @brief Program options not taking an argument. */
Option_no_arg opts_no_arg[NUM_OPTS_NO_ARG];

/** @brief Program options taking an argument. */
Option_with_arg opts_with_arg[NUM_OPTS_WITH_ARG];

/** @brief Director to put the generated training data. */
const char* data_out_dirname;


/** @brief Returns the number of options not taking an argument. */
uint32_t get_num_opts_no_arg()
{
    return NUM_OPTS_NO_ARG;
}

/** @brief Returns the number of options taking an argument. */
uint32_t get_num_opts_with_arg()
{
    return NUM_OPTS_WITH_ARG;
}

/** @brief Prints the program usage to stderr. */
void print_usage()
{
    fprintf(stderr, "usage: haplo-train OPTIONS [data-fname | <stdin>]\n");
    print_options(stderr, 27, NUM_OPTS_NO_ARG, opts_no_arg, NUM_OPTS_WITH_ARG,
            opts_with_arg);
}

/** @brief Process the 'data-out-dir' program option. */
Error* process_data_out_dir_opt(Option_arg arg)
{
    if (arg == NULL)
    {
        return JWSC_EARG("Option 'data-out-dir' requires an argument");
    }
    data_out_dirname = arg;

    return NULL;
}

/** @brief Initializes the haplo-train program options. */
static void init_train_options(void)
{
    uint32_t i;

    char s_name;
    const char* l_name;
    const char* desc;

    Error* (*farg)(const char*);

    init_options(opts_no_arg, opts_with_arg);

    opts.label_col = LABEL_COL;

    i = NUM_SHARED_OPTS_WITH_ARG;
    l_name = "model-dir";
    s_name = 0;
    desc   = "Directory to put trained models in.";
    farg   = process_model_dir_opt;
    init_option_with_arg(&(opts_with_arg[i++]), l_name, s_name, desc, farg);

    l_name = "data-out-dir";
    s_name = 0;
    desc   = "Directory to put the generated training data in for each model. The name of the model is used, so if this directory is set as the same as the model-dir, the models could be overwritten. The default is not to output the data.";
    farg   = process_data_out_dir_opt;
    init_option_with_arg(&(opts_with_arg[i++]), l_name, s_name, desc, farg);

    l_name = "nb-freq";
    s_name = 0;
    desc   = "Naive Bayes non-parametric frequency model tree information.";
    farg   = process_nb_freq_opt;
    init_option_with_arg(&(opts_with_arg[i++]), l_name, s_name, desc, farg);

    l_name = "nb-freq-dtd";
    s_name = 0;
    desc   = "Validate the naive Bayes non-parametric frequency model tree information XML file with this DTD.";
    farg   = process_nb_freq_dtd_opt;
    init_option_with_arg(&(opts_with_arg[i++]), l_name, s_name, desc, farg);

    l_name = "nb-gauss";
    s_name = 0;
    desc   = "Naive Bayes Gaussian model tree information.";
    farg   = process_nb_gauss_opt;
    init_option_with_arg(&(opts_with_arg[i++]), l_name, s_name, desc, farg);

    l_name = "nb-gauss-dtd";
    s_name = 0;
    desc   = "Validate the naive Bayes Gaussian model tree information XML file with this DTD.";
    farg   = process_nb_gauss_dtd_opt;
    init_option_with_arg(&(opts_with_arg[i++]), l_name, s_name, desc, farg);

    l_name = "nb-gmm";
    s_name = 0;
    desc   = "Naive Bayes Gaussian mixture model tree information.";
    farg   = process_nb_gmm_opt;
    init_option_with_arg(&(opts_with_arg[i++]), l_name, s_name, desc, farg);

    l_name = "nb-gmm-dtd";
    s_name = 0;
    desc   = "Validate the naive Bayes Gaussian mixture model tree information XML file with this DTD.";
    farg   = process_nb_gmm_dtd_opt;
    init_option_with_arg(&(opts_with_arg[i++]), l_name, s_name, desc, farg);

    l_name = "mv-gmm";
    s_name = 0;
    desc   = "Multivariate Gaussian mixture model tree information.";
    farg   = process_mv_gmm_opt;
    init_option_with_arg(&(opts_with_arg[i++]), l_name, s_name, desc, farg);

    l_name = "mv-gmm-dtd";
    s_name = 0;
    desc   = "Validate the multivariate Gaussian mixture model tree information XML file with this DTD.";
    farg   = process_mv_gmm_dtd_opt;
    init_option_with_arg(&(opts_with_arg[i++]), l_name, s_name, desc, farg);

    l_name = "mv-mmm";
    s_name = 0;
    desc   = "Multivariate multinomial mixture model tree information.";
    farg   = process_mv_mmm_opt;
    init_option_with_arg(&(opts_with_arg[i++]), l_name, s_name, desc, farg);

    l_name = "mv-mmm-dtd";
    s_name = 0;
    desc   = "Validate the multivariate multinomial mixture model tree information XML file with this DTD.";
    farg   = process_mv_mmm_dtd_opt;
    init_option_with_arg(&(opts_with_arg[i++]), l_name, s_name, desc, farg);

#ifdef HAPLO_ENABLE_SVM
    l_name = "svm";
    s_name = 0;
    desc   = "SVM model tree information.";
    farg   = process_svm_opt;
    init_option_with_arg(&(opts_with_arg[i++]), l_name, s_name, desc, farg);

    l_name = "svm-dtd";
    s_name = 0;
    desc   = "Validate the SVM model tree information XML file with this DTD.";
    farg   = process_svm_dtd_opt;
    init_option_with_arg(&(opts_with_arg[i++]), l_name, s_name, desc, farg);
#endif

#ifdef HAPLO_ENABLE_WEKA
    l_name = "weka-j48";
    s_name = 0;
    desc   = "Weka J48 model tree information.";
    farg   = process_weka_j48_opt;
    init_option_with_arg(&(opts_with_arg[i++]), l_name, s_name, desc, farg);

    l_name = "weka-part";
    s_name = 0;
    desc   = "Weka PART model tree information.";
    farg   = process_weka_part_opt;
    init_option_with_arg(&(opts_with_arg[i++]), l_name, s_name, desc, farg);

    l_name = "weka-jar";
    s_name = 0;
    desc   = "Weka java archive file. Required for using the Weka algorithms.";
    farg   = process_weka_jar_opt;
    init_option_with_arg(&(opts_with_arg[i++]), l_name, s_name, desc, farg);

    l_name = "weka-dtd";
    s_name = 0;
    desc   = "Validate the Weka model tree information XML files with this DTD.";
    farg   = process_weka_dtd_opt;
    init_option_with_arg(&(opts_with_arg[i++]), l_name, s_name, desc, farg);
#endif

    l_name = "nearest";
    s_name = 0;
    desc   = "Nearest neighbor model information.";
    farg   = process_nearest_opt;
    init_option_with_arg(&(opts_with_arg[i++]), l_name, s_name, desc, farg);

    l_name = "nearest-dtd";
    s_name = 0;
    desc   = "Validate the nearest neighbor model information XML file with this DTD.";
    farg   = process_nearest_dtd_opt;
    init_option_with_arg(&(opts_with_arg[i++]), l_name, s_name, desc, farg);
    assert(i == NUM_OPTS_WITH_ARG);
}


/** @brief Main function for the @em haplo-train program. */
int main(int argc, const char** argv)
{
    int         argi;
    const char* data_fname = "/dev/stdin";
    Error*      err;

    Matblock_u8* ids    = NULL;
    Vector_u32* labels  = NULL;
    Matrix_i32* markers = NULL;

    NB_freq_model_tree*  nb_freq_tree  = NULL;
    NB_gauss_model_tree* nb_gauss_tree = NULL;
    NB_gmm_model_tree*   nb_gmm_tree   = NULL;
    MV_gmm_model_tree*   mv_gmm_tree   = NULL;
    MV_mmm_model_tree*   mv_mmm_tree   = NULL;
#ifdef HAPLO_ENABLE_SVM
    SVM_model_tree*      svm_tree      = NULL;
#endif
#ifdef HAPLO_ENABLE_WEKA
    Weka_model_tree*     j48_tree      = NULL;
    Weka_model_tree*     part_tree     = NULL;
#endif
    Nearest_model*       nearest_model = NULL;

    init_train_options();

    if ((err = process_options(argc, argv, &argi, NUM_OPTS_NO_ARG, opts_no_arg,
                    NUM_OPTS_WITH_ARG, opts_with_arg)) != NULL)
    {
        print_error_msg_exit("haplo-train", err->msg);
    }

    if ((argc - argi) == 1)
    {
        data_fname = argv[ argi ];
    }

    if ((err = read_haplo_groups(opts.labels_fname)))
    {
        print_error_msg_exit("haplo-train", err->msg);
    }

    if ((err = read_input(&ids, &labels, &markers, data_fname)))
    {
        print_error_msg_exit("haplo-train", err->msg);
    }

    if (!labels)
    {
        print_error_msg("haplo-train", NULL);
        print_error_msg_exit(data_fname, "No labels to train with");
    }

    if (opts.nb_freq_fname)
    {
        if ((err = train_nb_freq_model_tree(&nb_freq_tree, labels, markers,
                        opts.nb_freq_fname, opts.nb_freq_dtd_fname)) ||
            (err = write_nb_freq_model_tree(nb_freq_tree, opts.model_dirname)))
        {
            print_error_msg_exit("haplo-train", err->msg);
        }
        free_nb_freq_model_tree(nb_freq_tree);
    }

    if (opts.nb_gauss_fname)
    {
        if ((err = train_nb_gauss_model_tree(&nb_gauss_tree, labels, markers, 
                        opts.nb_gauss_fname, opts.nb_gauss_dtd_fname)) ||
            (err = write_nb_gauss_model_tree(nb_gauss_tree, 
                        opts.model_dirname)))
        {
            print_error_msg_exit("heplo-train", err->msg);
        }
        free_nb_gauss_model_tree(nb_gauss_tree);
    }

    if (opts.nb_gmm_fname)
    {
        if ((err = train_nb_gmm_model_tree(&nb_gmm_tree, labels, markers, 
                        opts.nb_gmm_fname, opts.nb_gmm_dtd_fname)) ||
            (err = write_nb_gmm_model_tree(nb_gmm_tree, opts.model_dirname)))
        {
            print_error_msg_exit("haplo-train", err->msg);
        }
        free_nb_gmm_model_tree(nb_gmm_tree);
    }

    if (opts.mv_gmm_fname)
    {
        if ((err = train_mv_gmm_model_tree(&mv_gmm_tree, labels, markers, 
                        opts.mv_gmm_fname, opts.mv_gmm_dtd_fname)) ||
            (err = write_mv_gmm_model_tree(mv_gmm_tree, opts.model_dirname)))
        {
            print_error_msg_exit("haplo-train", err->msg);
        }
        free_mv_gmm_model_tree(mv_gmm_tree);
    }

    if (opts.mv_mmm_fname)
    {
        if ((err = train_mv_mmm_model_tree(&mv_mmm_tree, labels, markers, 
                        opts.mv_mmm_fname, opts.mv_mmm_dtd_fname)) ||
            (err = write_mv_mmm_model_tree(mv_mmm_tree, opts.model_dirname)))
        {
            print_error_msg_exit("haplo-train", err->msg);
        }
        free_mv_mmm_model_tree(mv_mmm_tree);
    }

#ifdef HAPLO_ENABLE_SVM
    if (opts.svm_fname)
    {
        if ((err = train_svm_model_tree(&svm_tree, labels, markers,
                        opts.svm_fname, opts.svm_dtd_fname)) ||
            (err = write_svm_model_tree(svm_tree, opts.model_dirname)))
        {
            print_error_msg_exit("heplo-train", err->msg);
        }
        free_svm_model_tree(svm_tree);
        if (data_out_dirname)
        {
            if ((err = write_svm_model_tree_training_data(labels, markers,
                            opts.svm_fname, opts.svm_dtd_fname,
                            data_out_dirname)))
            {
                print_error_msg_exit("haplo-train", err->msg);
            }
        }
    }
#endif

#ifdef HAPLO_ENABLE_WEKA
    if (opts.weka_j48_fname)
    {
        if ((err = train_weka_j48_model_tree(&j48_tree, labels, markers,
                        opts.weka_j48_fname, opts.weka_dtd_fname,
                        opts.model_dirname, opts.weka_jar_fname)))
        {
            print_error_msg_exit("heplo-train", err->msg);
        }
        free_weka_model_tree(j48_tree);
    }

    if (opts.weka_part_fname)
    {
        if ((err = train_weka_part_model_tree(&part_tree, labels, markers,
                        opts.weka_part_fname, opts.weka_dtd_fname,
                        opts.model_dirname, opts.weka_jar_fname)))
        {
            print_error_msg_exit("heplo-train", err->msg);
        }
        free_weka_model_tree(part_tree);
    }
#endif

    if (opts.nearest_fname)
    {
        if ((err = train_nearest_model(&nearest_model, labels, markers,
                        opts.nearest_fname, opts.nearest_dtd_fname)) ||
            (err = write_nearest_model(nearest_model, opts.model_dirname)))
        {
            print_error_msg_exit("haplo-train", err->msg);
        }
        free_nearest_model(nearest_model);
    }

    free_matblock_u8(ids);
    free_vector_u32(labels);
    free_matrix_i32(markers);

    if (get_num_unhandled_errors() > 0)
    {
        print_error_msg_exit("haplo-train", "Unhandled errors");
    }

    return EXIT_SUCCESS;
}
