/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Read the a row-by-row collection of data samples from a file.
 */


#ifndef INPUT_H
#define INPUT_H


#include <config.h>

#include <stdlib.h>
#include <inttypes.h>

#include <jwsc/vector/vector.h>
#include <jwsc/matrix/matrix.h>
#include <jwsc/matblock/matblock.h>


/** @brief Types of input file format to read. */
typedef enum
{
    HAPLO_INPUT_TXT,
    HAPLO_INPUT_CSV,
    HAPLO_INPUT_XML
}
Haplo_input_format;


/** @brief Reads the IDs, haplo labels and markers from a file. */
Error* read_input
(
    Matblock_u8** ids_out,
    Vector_u32**  labels_out,
    Matrix_i32**  markers_out,
    const char*   fname
);


/** @brief Reads the IDs, haplo labels and markers from an auxiliary file. */
Error* read_aux_input
(
    Matblock_u8** ids_out,
    Vector_u32**  labels_out,
    Matrix_i32**  markers_out,
    const char*   fname
);


/** 
 * @brief Imputes missing marker values (zero) with their label-dependent
 * mean value.
 */
Error* impute_missing_markers_from_avg
(
    const Vector_u32* imp_labels, 
    Matrix_i32*       imp_markers,
    const Vector_u32* src_labels, 
    const Matrix_i32* src_markers
);


/** @brief imputes a given marker with the values from the parent haplogroup
	this happens when we have only one sample with a given haplogroup */

Error* impute_from_parent_of_haplogroup_index(
	uint32_t haplo_group_index,
	uint32_t marker_no,
	uint32_t sample_no,
	Matrix_i32* marker_sums,
	Matrix_i32* imp_markers
	
	);



/** 
 * @brief Imputes missing marker values (zero) from nearest neighbor
 * interpolation.
 */
Error* impute_missing_markers_from_nn
(
    Matrix_i32*       imp_markers,
    const Matrix_i32* src_markers
);


#endif
