/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Naive Bayes non-parametric marker frequency classifier.
 */


#ifndef NB_FREQ_H
#define NB_FREQ_H


#include <config.h>

#include <stdlib.h>
#include <inttypes.h>

#include <jwsc/base/error.h>
#include <jwsc/vector/vector.h>
#include <jwsc/matrix/matrix.h>


/** @brief Naive Bayes non-parametric marker frequency model. */
typedef struct
{
    /** @brief Number of markers. */
    uint32_t  num_markers;

    /** @brief Minimum marker value. */
    uint32_t  min_marker_val;

    /** @brief Number of integer marker values. */
    uint32_t  num_marker_vals;

    /** @brief Likelihood table for the ith marker given a label. */
    Matrix_d* marker_given_label;

    /** 
     * @brief Prior probabilities over the labels.
     *
     * Priors are either calculated from the frequency of training data or
     * supplied by the user.
     */
    Vector_d* priors;
} 
NB_freq_model;


/** @brief Trains a Naive Bayes non-parametric marker frequency model. */
void train_nb_freq_model
(
    NB_freq_model**   model_out,
    const Vector_u32* labels, 
    const Matrix_i32* markers,
    const Vector_d*   priors
);


/** 
 * @brief Predicts the label for a marker sample using a Naive Bayes 
 * non-parametric marker frequency model.
 */
Error* predict_label_with_nb_freq_model
(
    uint32_t*            label_out,
    double*              confidence_out,
    const Vector_i32*    markers,
    const NB_freq_model* model,
    uint32_t             order
);


/** 
 * @brief Predicts the labels for a set of marker samples using a Naive
 * Bayes non-parametric marker frequency model.
 */
Error* predict_labels_with_nb_freq_model
(
    Vector_u32**         labels_out,
    Vector_d**           confidence_out,
    const Matrix_i32*    markers,
    const NB_freq_model* model
);


/** @brief Reads a Naive Bayes non-parametric marker frequency model. */
Error* read_nb_freq_model(NB_freq_model** model_out, const char* fname);


/** @brief Writes a Naive Bayes non-parametric marker frequency model. */
Error* write_nb_freq_model(NB_freq_model* model, const char* fname);


/** @brief Frees a Naive Bayes non-parametric marker frequency model. */
void free_nb_freq_model(NB_freq_model* model);


/** @brief Naive Bayes non-parametric marker frequency model tree node. */
typedef struct NB_freq_model_node
{
    /** 
     * @brief Parent of a model node.
     *
     * NULL if the node is the root of the model tree.
     */
    const struct NB_freq_model_node* parent;

    /** 
     * @brief Parent label for a model node.
     *
     * If the parent is NULL, the value is undefined.
     */
    uint32_t parent_label;

    /** @brief Number of model groups in a node. */
    uint32_t num_groups;

    /** 
     * @brief Model subtrees for each group in a model node.
     * 
     * The ith element is NULL if the ith group does not have a subtree.
     */
    struct NB_freq_model_node** subtrees;

    /** @brief Labels for each group in a model node. */
    Vector_u32* labels;

    /** @brief Alternative label sets for each group in a model node. */
    Vector_u32** altlabels;

    /** 
     * @brief Label priors for each group in a model node supplied by the user. 
     *
     * If the priors were not supplied by the user, set to NULL.
     */
    Vector_d* priors;

    /** @brief Naive Bayes marker frequency model. */
    NB_freq_model* model;

    /** @brief File name for storage of the model at a node. */
    char* model_fname;
} 
NB_freq_model_node;


/** @brief Naive Bayes non-parametric marker frequency model tree. */
typedef  NB_freq_model_node  NB_freq_model_tree;


/** @brief Trains a Naive Bayes non-parametric marker frequency model tree . */
Error* train_nb_freq_model_tree
(
    NB_freq_model_tree** tree_out,
    const Vector_u32*    labels, 
    const Matrix_i32*    markers,
    const char*          tree_xml_fname,
    const char*          tree_dtd_fname
);


/** 
 * @brief Predicts the labels for a set of marker samples using a Naive Bayes
 * non-parametric marker frequency model tree.
 */
Error* predict_labels_with_nb_freq_model_tree
(
    Vector_u32**              labels_out,
    Vector_d**                confidence_out,
    const Matrix_i32*         markers,
    const NB_freq_model_tree* tree,
    uint32_t                  order
);


/**
 * @brief Reads a Naive Bayes non-parametric marker frequency model tree 
 * from.
 */
Error* read_nb_freq_model_tree
(
    NB_freq_model_tree** tree_out,
    const char*          tree_xml_fname,
    const char*          tree_dtd_fname,
    const char*          model_dirname
);


/** @brief Writes a Naive Bayes non-parametric marker frequency model tree. */
Error* write_nb_freq_model_tree
(
    const NB_freq_model_tree* tree,
    const char*               model_dirname
);


/** @brief Frees a Naive Bayes non-parametric marker frequency model tree. */
void free_nb_freq_model_tree(NB_freq_model_tree* tree);


#endif
