/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Nearest neighbor classifier.
 */


#include <config.h>

#include <stdlib.h>
#include <stdio.h>
#include <inttypes.h>
#include <assert.h>
#include <string.h>
#include <errno.h>

#include <libxml/tree.h>
#include <libxml/parser.h>
#include <libxml/valid.h>

#ifdef HAPLO_HAVE_DMALLOC
#include <dmalloc.h>
#endif

#include <jwsc/base/error.h>
#include <jwsc/base/limits.h>
#include <jwsc/vector/vector.h>
#include <jwsc/vector/vector_io.h>
#include <jwsc/matrix/matrix.h>
#include <jwsc/matrix/matrix_io.h>

#include "xml.h"
#include "haplo_groups.h"
#include "nearest.h"


/** @brief Reads and and optionally validates an XML document. */
static Error* read_nearest_xml_doc
(
    xmlDoc**    xml_doc_out,
    const char* xml_fname,
    const char* dtd_fname
)
{
    xmlParserCtxt* xml_parse_ctxt;
    xmlValidCtxt*  xml_valid_ctxt;
    xmlDtd*        xml_dtd;
    int            slen = 256;
    char           str[slen];

    assert(xml_parse_ctxt = xmlNewParserCtxt());

    if (!(*xml_doc_out = xmlCtxtReadFile(xml_parse_ctxt, xml_fname, NULL, 0)))
    {
        snprintf(str, slen, "%s: %s", xml_fname, "Could not parse file");
        return JWSC_EARG(str);
    } 

    xmlFreeParserCtxt(xml_parse_ctxt);

    if (dtd_fname)
    {
        assert(xml_valid_ctxt = xmlNewValidCtxt());

        if (!(xml_dtd = xmlParseDTD(NULL, (xmlChar*)dtd_fname)))
        {
            snprintf(str, slen, "%s: %s", dtd_fname, "Could not parse DTD");
            return JWSC_EARG(str);
        }

        if (!xmlValidateDtd(xml_valid_ctxt, *xml_doc_out, xml_dtd))
        {
            snprintf(str, slen, "%s: %s", xml_fname, "XML file not valid");
            return JWSC_EARG(str);
        }

        xmlFreeValidCtxt(xml_valid_ctxt);
        xmlFreeDtd(xml_dtd);
    }

    return NULL;
}


/**
 * @param  model_out        Result parameter. If @em *model_out is NULL, a
 *                          model is allocated; otherwise its space is re-used.
 * @param  labels           Labels for training, with ith element as 
 *                          corresponding to the ith sample in @em markers.
 * @param  markers          Markers for training, with ith row as a sample 
 *                          corresponding to the ith label in @em labels.
 * @param  model_xml_fname  XML file containing the model information.
 * @param  model_dtd_fname  DTD file for validating the XML file, can be NULL.
 */
Error* train_nearest_model
(
    Nearest_model**   model_out,
    const Vector_u32* labels, 
    const Matrix_i32* markers,
    const char*       model_xml_fname,
    const char*       model_dtd_fname
)
{
    Nearest_model* model;
    xmlDoc*     xml_doc;
    xmlNode*    it;
    Error*      err;
    const char* fname;

    assert(model_xml_fname);
    assert(markers->num_rows == labels->num_elts);

    if (*model_out != NULL)
    {
        free_nearest_model(*model_out);
    }
    assert(*model_out = malloc(sizeof(Nearest_model)));
    model = *model_out;

    model->labels = NULL;
    model->markers = NULL;
    model->fname = NULL;

    copy_vector_u32(&(model->labels), labels);
    copy_matrix_i32(&(model->markers), markers);

    if ((err = read_nearest_xml_doc(&xml_doc, model_xml_fname,
                    model_dtd_fname)))
    {
        return err;
    }

    for (it = xmlDocGetRootElement(xml_doc)->children; it; it = it->next)
    {
        if (it->type == XML_ELEMENT_NODE && XMLStrEqual(it->name, "file"))
        {
            fname = (const char*) it->children->content;
            model->fname = malloc((strlen(fname)+1) * sizeof(char));
            strcpy(model->fname, fname);
            break;
        }
    }

    return NULL;
}


/** 
 * @param  label_out       Result parameter. 
 * @param  dist_out        Result parameter. 
 * @param  markers         Marker data to predict.
 * @param  model           Trained model to use for predicting.
 *
 * @return On success, NULL is returned; otherwise an error is returned.
 */
Error* predict_label_with_nearest_model
(
    uint32_t*            label_out,
    double*              dist_out,
    const Vector_i32*    markers,
    const Nearest_model* model
)
{
    uint32_t label, num_labels;
    uint32_t s, num_samples;
    uint32_t m, num_markers;
    uint32_t d, D;

    num_labels  = get_num_haplo_groups();
    num_samples = model->markers->num_rows;
    num_markers = markers->num_elts;

    assert(num_markers == model->markers->num_cols);

    label = 0;
    D     = JWSC_MAX_INT32;

    for (s = 0; s < num_samples; s++)
    {
        d = 0;

        for(m = 0; m < num_markers; m++)
        {
            d += abs(markers->elts[ m ] - model->markers->elts[ s ][ m ]);
        }

        if (d < D)
        {
            label = model->labels->elts[ s ];
            D = d;
        }
        else if (d == D)
        { 
            if (is_ancestor(model->labels->elts[ s ], label))
            {
                label = model->labels->elts[ s ];
            }
        }
    }

    *label_out = label;
    *dist_out  = (double)D;

    return NULL;
}


/** 
 * @param  labels_out      Result parameter. If @em *labels_out is NULL, it is
 *                         allocated; otherwise its space is re-used.
 * @param  dists_out       Result parameter. If @em *dists_out is NULL, it is
 *                         allocated; otherwise its space is re-used.
 * @param  markers         Marker data to predict. Each row is a sample for
 *                         prediction, corresponding to an element in the result
 *                         parameters.
 * @param  model           Trained model to use for predicting.
 *
 * @return On success, NULL is returned; otherwise an error is returned, but the
 * result parameters are not freed.
 */
Error* predict_labels_with_nearest_model
(
    Vector_u32**         labels_out,
    Vector_d**           dists_out,
    const Matrix_i32*    markers,
    const Nearest_model* model
)
{
    uint32_t s, num_samples;
    uint32_t m, num_markers;

    Vector_u32* labels;
    Vector_d*   dists;
    Vector_i32* markers_v;
    Error*      e;

    num_samples = markers->num_rows;
    num_markers = markers->num_cols;

    create_vector_u32(labels_out, num_samples);
    labels = *labels_out;

    create_vector_d(dists_out, num_samples);
    dists = *dists_out;

    markers_v = NULL;
    create_vector_i32(&markers_v, num_markers);

    for (s = 0; s < num_samples; s++)
    {
        for(m = 0; m < num_markers; m++)
        {
            markers_v->elts[ m ] = markers->elts[ s ][ m ];
        }

        if ((e = predict_label_with_nearest_model(&(labels->elts[ s ]),
                        &(dists->elts[ s ]), markers_v, model)) != NULL)
        {
            free_vector_i32(markers_v);
            return e;
        }
    }

    free_vector_i32(markers_v);

    return NULL;
}


/** @brief Reads the nearest neighbor model. */
Error* read_nearest_model
(
    Nearest_model** model_out,
    const char*     model_xml_fname,
    const char*     model_dtd_fname,
    const char*     model_dirname
)
{
    Nearest_model* model;
    xmlDoc*   xml_doc;
    xmlNode*  it;
    FILE*     fp;
    Error*    e;
    char      buf[1024] = {0};
    const char* fname;
    int       slen = 256;
    char      str[slen];

    if (*model_out != NULL)
    {
        free_nearest_model(*model_out);
    }
    assert(*model_out = malloc(sizeof(Nearest_model)));
    model = *model_out;

    model->labels = NULL;
    model->markers = NULL;
    model->fname = NULL;

    if ((e = read_nearest_xml_doc(&xml_doc, model_xml_fname, model_dtd_fname)))
    {
        return e;
    }

    for (it = xmlDocGetRootElement(xml_doc)->children; it; it = it->next)
    {
        if (it->type == XML_ELEMENT_NODE && XMLStrEqual(it->name, "file"))
        {
            fname = (const char*) it->children->content;
            model->fname = malloc((strlen(fname)+1) * sizeof(char));
            strcpy(model->fname, fname);
            break;
        }
    }

    snprintf(buf, 1024, "%s/%s", model_dirname, model->fname);

    if ((fp = fopen(buf, "r")) == NULL)
    {
        snprintf(str, slen, "%s: %s", buf, strerror(errno));
        return JWSC_EARG(str);
    }

    if ((e = read_vector_with_header_fp_u32(&(model->labels), fp)))
    {
        snprintf(str, slen, "%s: %s", buf, e->msg);
        return JWSC_EARG(str);
    }

    if ((e = read_matrix_with_header_fp_i32(&(model->markers), fp)))
    {
        snprintf(str, slen, "%s: %s", buf, e->msg);
        return JWSC_EARG(str);
    }

    if (fclose(fp) != 0)
    {
        snprintf(str, slen, "%s: %s", buf, strerror(errno));
        return JWSC_EARG(str);
    }

    return NULL;
}


/** @brief Writes the nearest neighbor model. */
Error* write_nearest_model
(
    const Nearest_model* model,
    const char*          model_dirname
)
{
    char    buf[1024] = {0};
    FILE*   fp;
    Error*  e;
    int     slen = 256;
    char    str[slen];

    snprintf(buf, 1024, "%s/%s", model_dirname, model->fname);

    if ((fp = fopen(buf, "w")) == NULL)
    {
        snprintf(str, slen, "%s: %s", buf, strerror(errno));
        return JWSC_EARG(str);
    }

    if ((e = write_vector_with_header_fp_u32(model->labels, fp)) != NULL)
    {
        snprintf(str, slen, "%s: %s", buf, e->msg);
        return JWSC_EARG(str);
    }

    if ((e = write_matrix_with_header_fp_i32(model->markers, fp)) != NULL)
    {
        snprintf(str, slen, "%s: %s", buf, e->msg);
        return JWSC_EARG(str);
    }

    if (fclose(fp) != 0)
    {
        snprintf(str, slen, "%s: %s", buf, strerror(errno));
        return JWSC_EARG(str);
    }

    return NULL;
}


/** @param  model  Model to free. */
void free_nearest_model(Nearest_model* model)
{
    if (!model)
        return;

    free_vector_u32(model->labels);
    free_matrix_i32(model->markers);
    free(model->fname);
    free(model);
}
