/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Nearest neighbor classifier.
 */


#ifndef NEAREST_H
#define NEAREST_H


#include <config.h>

#include <stdlib.h>
#include <inttypes.h>

#include <jwsc/base/error.h>
#include <jwsc/vector/vector.h>
#include <jwsc/matrix/matrix.h>


/** @brief Nearest neighbor model (the whole data-set). */
typedef struct
{
    /** @brief Label for each sample. */
    Vector_u32* labels;

    /** @brief Marker data. */
    Matrix_i32* markers;

    /** @brief Model file name. */
    char* fname;
} 
Nearest_model;


/** @brief Trains a Naive Bayes non-parametric marker frequency model. */
Error* train_nearest_model
(
    Nearest_model**   model_out,
    const Vector_u32* labels, 
    const Matrix_i32* markers,
    const char*       model_xml_fname,
    const char*       model_dtd_fname
);


/** @brief Predicts a label for a marker sample using a nearest neighbor. */
Error* predict_label_with_nearest_model
(
    uint32_t*            label_out,
    double*              dist_out,
    const Vector_i32*    markers,
    const Nearest_model* model
);


/** 
 * @brief Predicts the labels for a set of marker samples using nearest
 * neighbor.
 */
Error* predict_labels_with_nearest_model
(
    Vector_u32**         labels_out,
    Vector_d**           dists_out,
    const Matrix_i32*    markers,
    const Nearest_model* model
);


/** @brief Reads the nearest neighbor model. */
Error* read_nearest_model
(
    Nearest_model** model_out,
    const char*     model_xml_fname,
    const char*     model_dtd_fname,
    const char*     model_dirname
);


/** @brief Writes the nearest neighbor model. */
Error* write_nearest_model
(
    const Nearest_model* model,
    const char*          model_dirname
);


/** @brief Frees a nearest neighbor model. */
void free_nearest_model(Nearest_model* model);


#endif
