/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Options used for all haplo programs.
 */


#ifndef OPTIONS_H
#define OPTIONS_H


#include <config.h>

#include <stdlib.h>
#include <stdio.h>
#include <inttypes.h>

#include <jwsc/base/error.h>
#include <jwsc/base/option.h>
#include <jwsc/vector/vector.h>

#include "output.h"
#include "input.h"


#ifdef HAPLO_HAVE_PTHREAD
#define NUM_THREAD_OPTS 1
#else
#define NUM_THREAD_OPTS 0
#endif

#define  NUM_SHARED_OPTS_NO_ARG    4
#define  NUM_SHARED_OPTS_WITH_ARG  12 + NUM_THREAD_OPTS


/** @brief Haplo program shared option values. */
struct
{
    /** 
     * @brief Boolean indicating whether the input data has a header
     * (descriptive) line.
     */
    uint8_t header_in;

    /** 
     * @brief Boolean indicating whether to write a header (descriptive) line 
     * to the output file.
     */
    uint8_t header_out;

    /** 
     * @brief Boolean indicating whether to exclude at most one prediction
     * when deciding on an agreed label from the set of classification
     * algorithms.
     */
    uint8_t exclude_one;

    /** 
     * @brief Random seed.
     *
     * The random generator is initially seeded (srand) with zero.
     */
    uint32_t seed;

    /** @brief Number of pthreads to use at any given time. */
    uint32_t num_threads;

    /** 
     * @brief Input file format.
     *
     * Current choices include text, comman separated values, or xml.
     */
    Haplo_input_format input_format;

    /** @brief XML input file DTD. */
    const char* input_dtd_fname;

    /** 
     * @brief Output file format.
     *
     * Current choices include text, comman separated values, or xml.
     */
    Haplo_output_format output_format;

    /** @brief XML file containin the haplo group labels. */
    const char* labels_fname;

    /** @brief XML labels file DTD. */
    const char* labels_dtd_fname;

    /** @brief Columns in input CSV file containing sample IDs. */
    Vector_u32* id_cols;

    /** @brief Column in input CSV file containing the haplo labels. */
    uint32_t label_col;

    /** @brief Column in input CSV file containing the first marker. */
    uint32_t first_marker_col;

    /** @brief Number of marker columns in the input CSV file. */
    uint32_t num_markers;

    /** @brief Marker columns to use. */
    Vector_u32* marker_cols;

    /** @brief Auxiliary input file. */
    const char* aux_input_fname;

    /** @brief Columns in input CSV file containing sample IDs. */
    Vector_u32* aux_id_cols;

    /** @brief Column in input CSV file containing the haplo labels. */
    uint32_t aux_label_col;

    /** @brief Column in input CSV file containing the first marker. */
    uint32_t aux_first_marker_col;

    /** @brief Number of marker columns in the input CSV file. */
    uint32_t aux_num_markers;

    /** @brief Marker columns to use. */
    Vector_u32* aux_marker_cols;

    /** @brief Number of marker columns in the input CSV file. */
    uint32_t num_gmm_markers;

    /** @brief Naive Bayes non-parametric frequency model tree information. */
    const char* nb_freq_fname;

    /** @brief XML DTD for nb_freq_fname. */
    const char* nb_freq_dtd_fname;

    /** @brief Naive Bayes Gaussian model tree information. */
    const char* nb_gauss_fname;

    /** @brief XML DTD for nb_gauss_fname. */
    const char* nb_gauss_dtd_fname;

    /** @brief Naive Bayes Gaussian mixture model tree information. */
    const char* nb_gmm_fname;

    /** @brief XML DTD for nb_gmm_fname. */
    const char* nb_gmm_dtd_fname;

    /** @brief Multivariate Gaussian mixture model tree information. */
    const char* mv_gmm_fname;

    /** @brief XML DTD for mv_gmm_fname. */
    const char* mv_gmm_dtd_fname;

    /** @brief Multivariate multinomial mixture model tree information. */
    const char* mv_mmm_fname;

    /** @brief XML DTD for mv_mmm_fname. */
    const char* mv_mmm_dtd_fname;

#ifdef HAPLO_ENABLE_SVM
    /** @brief SVM model tree information. */
    const char* svm_fname;

    /** @brief XML DTD for svm_fname. */
    const char* svm_dtd_fname;
#endif

#ifdef HAPLO_ENABLE_WEKA
    /** @brief Weka J48 model tree information. */
    const char* weka_j48_fname;

    /** @brief Weka PART model tree information. */
    const char* weka_part_fname;

    /** @brief Weka java archive file. */
    const char* weka_jar_fname;

    /** @brief XML DTD for Weka models. */
    const char* weka_dtd_fname;
#endif

    /** @brief Nearest neighbor model information. */
    const char* nearest_fname;

    /** @brief XML DTD for the nearest neighbor model. */
    const char* nearest_dtd_fname;

    /** @brief Maximum allowed nearest neighbor distance for a prediction. */
    uint32_t nearest_max_d;

    /** @brief Name of the Bernoulli mixture model tree file. */
    const char* bmm_tree_fname;

    /** @brief Name of the Bernoulli mixture model info file. */
    const char* bmm_info_fname;

    /** @brief Name of the Bernoulli mixture model confidence threshold file. */
    const char* bmm_conf_thresh_fname;

    /** @brief Name of the Gaussian and Bernoulli mixture model tree file. */
    const char* gmm_bmm_tree_fname;

    /** @brief Name of the Gaussian and Bernoulli mixture model info file. */
    const char* gmm_bmm_info_fname;

    /** 
     * @brief Name of the Gaussian and Bernoulli mixture model confidence
     * threshold file.
     */
    const char* gmm_bmm_conf_thresh_fname;

    /** @brief Director to find the models in. */
    const char* model_dirname;
}
opts;

/** @brief Process the 'exclude-one' program option. */
Error* process_exclude_one_opt();

/** @brief Process the 'aux-input' program option. */
Error* process_aux_input_opt(Option_arg arg);

/** @brief Process the 'aux-id-cols' program option. */
Error* process_aux_id_cols_opt(Option_arg arg);

/** @brief Process the 'aux-label-col' program option. */
Error* process_aux_label_col_opt(Option_arg arg);

/** @brief Process the 'aux-1st-marker-col' program option. */
Error* process_aux_first_marker_col_opt(Option_arg arg);

/** @brief Process the 'aux-num-markers' program option. */
Error* process_aux_num_markers_opt(Option_arg arg);

/** @brief Process the 'aux-marker-cols' program option. */
Error* process_aux_marker_cols_opt(Option_arg arg);

/** @brief Process the 'num-gmm-bmm-markers' program option. */
Error* process_num_gmm_bmm_markers_opt(Option_arg arg);

/** @brief Process the 'model-dir' program option. */
Error* process_model_dir_opt(Option_arg arg);

/** @brief Process the 'nb-freq' program option. */
Error* process_nb_freq_opt(Option_arg arg);

/** @brief Process the 'nb-freq-dtd' program option. */
Error* process_nb_freq_dtd_opt(Option_arg arg);

/** @brief Process the 'nb-gauss' program option. */
Error* process_nb_gauss_opt(Option_arg arg);

/** @brief Process the 'nb-gauss-dtd' program option. */
Error* process_nb_gauss_dtd_opt(Option_arg arg);

/** @brief Process the 'nb-gmm' program option. */
Error* process_nb_gmm_opt(Option_arg arg);

/** @brief Process the 'nb-gmm-dtd' program option. */
Error* process_nb_gmm_dtd_opt(Option_arg arg);

/** @brief Process the 'mv-gmm' program option. */
Error* process_mv_gmm_opt(Option_arg arg);

/** @brief Process the 'mv-gmm-dtd' program option. */
Error* process_mv_gmm_dtd_opt(Option_arg arg);

/** @brief Process the 'mv-mmm' program option. */
Error* process_mv_mmm_opt(Option_arg arg);

/** @brief Process the 'mv-mmm-dtd' program option. */
Error* process_mv_mmm_dtd_opt(Option_arg arg);

#ifdef HAPLO_ENABLE_SVM
/** @brief Process the 'svm' program option. */
Error* process_svm_opt(Option_arg arg);

/** @brief Process the 'svm-dtd' program option. */
Error* process_svm_dtd_opt(Option_arg arg);
#endif

#ifdef HAPLO_ENABLE_WEKA
/** @brief Process the 'weka-j48' program option. */
Error* process_weka_j48_opt(Option_arg arg);

/** @brief Process the 'weka-part' program option. */
Error* process_weka_part_opt(Option_arg arg);

/** @brief Process the 'weka-jar' program option. */
Error* process_weka_jar_opt(Option_arg arg);

/** @brief Process the 'weka-dtd' program option. */
Error* process_weka_dtd_opt(Option_arg arg);
#endif

/** @brief Process the 'nearest' program option. */
Error* process_nearest_opt(Option_arg arg);

/** @brief Process the 'nearest-dtd' program option. */
Error* process_nearest_dtd_opt(Option_arg arg);

/** @brief Process the 'nearest-max-d' program option. */
Error* process_nearest_max_d_opt(Option_arg arg);

/** @brief Process the 'bmm-tree' program option. */
Error* process_bmm_tree_opt(Option_arg arg);

/** @brief Process the 'bmm-info' program option. */
Error* process_bmm_info_opt(Option_arg arg);

/** @brief Process the 'bmm-conf-thresh' program option. */
Error* process_bmm_conf_thresh_opt(Option_arg arg);

/** @brief Process the 'gmm-bmm-tree' program option. */
Error* process_gmm_bmm_tree_opt(Option_arg arg);

/** @brief Process the 'gmm-bmm-info' program option. */
Error* process_gmm_bmm_info_opt(Option_arg arg);

/** @brief Process the 'gmm-bmm-conf-thresh' program option. */
Error* process_gmm_bmm_conf_thresh_opt(Option_arg arg);

/** @brief Initializes the shared program options. */
void init_options(Option_no_arg* opts_no_arg, Option_with_arg* opts_with_arg);


#endif
