/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Support vector machine model classifier.
 */


#ifndef HAPLO_SVM_H
#define HAPLO_SVM_H


#include <config.h>


#include <stdlib.h>
#include <inttypes.h>

#if defined HAPLO_HAVE_LIBSVM_H
#include <libsvm.h>
#elif defined HAPLO_HAVE_SVM_H
#include <svm.h>
#endif

#include <jwsc/base/error.h>
#include <jwsc/vector/vector.h>
#include <jwsc/matrix/matrix.h>


/** @brief Support vector machine model. */
typedef struct
{
    /** @brief libSVM model data structure. */
    struct svm_model* svm;

    /** 
     * @brief Requisite libSVM problem created for training.
     *
     * when the model is loaded from a file, this field is NULL.
     */
    struct svm_problem* prob;
}
SVM_model;


/** @brief Trains a SVM model. */
void train_svm_model
(
    SVM_model**       model_out,
    const Vector_u32* labels, 
    const Matrix_i32* markers,
    double            cost,
    double            gamma
);


/** @brief Predicts the label for a marker sample using a SVM model. */
Error* predict_label_with_svm_model
(
    uint32_t*         label_out,
    double**          confidence_out,
    const struct svm_node*  markers,
    const SVM_model*  model
);


/** @brief Predicts the labels for a set of marker samples using a SVM model. */
Error* predict_labels_with_svm_model
(
    Vector_u32**      labels_out,
    Vector_d**        confidence_out,
    const Matrix_i32* markers,
    const SVM_model*  model
);


/** @brief Reads a SVM model. */
Error* read_svm_model(SVM_model** model_out, const char* fname);


/** @brief Writes a SVM model. */
Error* write_svm_model(SVM_model* model, const char* fname);


/** @brief Writes the training data for an SVM model to a file. */
Error* write_svm_model_training_data
(
    const Vector_u32* labels,
    const Matrix_i32* markers,
    const char*       fname
);


/** @brief Frees a SVM model. */
void free_svm_model(SVM_model* model);


/** @brief SVM model tree node. */
typedef struct SVM_model_node
{
    /** 
     * @brief Parent of a model node.
     *
     * NULL if the node is the root of the model tree.
     */
    const struct SVM_model_node* parent;

    /** 
     * @brief Parent label for a model node.
     *
     * If the parent is NULL, the value is undefined.
     */
    uint32_t parent_label;

    /** 
     * @brief Subtrees for each model in a node.
     * 
     * The ith element is NULL if the ith group does not have a subtree.
     */
    struct SVM_model_node** subtrees[2];

    /** @brief Labels for the two groups in a model node. */
    Vector_u32* labels[2];

    /** @brief Alternative label sets for the two groups in a model node. */
    Vector_u32** altlabels[2];

    /** @brief Cost parameter for each model node. */
    Vector_d* cost;

    /** @brief Gamma parameter for each model node. */
    Vector_d* gamma;

    /** @brief Number of SVM models. */
    uint32_t num_models; 

    /** @brief SVM model. */
    SVM_model** models;

    /** @brief File name for storage of the model at a node. */
    char** model_fnames;
} 
SVM_model_node;


/** @brief SVM model tree. */
typedef  SVM_model_node  SVM_model_tree;


/** @brief Trains a SVM model tree . */
Error* train_svm_model_tree
(
    SVM_model_tree**    tree_out,
    const Vector_u32*   labels, 
    const Matrix_i32*   markers,
    const char*         tree_xml_fname,
    const char*         tree_dtd_fname
);


/** 
 * @brief Predicts the labels for a set of marker samples using a SVM model
 * tree.
 */
Error* predict_labels_with_svm_model_tree
(
    Vector_u32**          labels_out,
    Vector_d**            confidence_out,
    const Matrix_i32*     markers,
    const SVM_model_tree* tree
);


/** @brief Reads a SVM model tree from. */
Error* read_svm_model_tree
(
    SVM_model_tree** tree_out,
    const char*      tree_xml_fname,
    const char*      tree_dtd_fname,
    const char*      model_dirname
);


/** @brief Writes a SVM model tree. */
Error* write_svm_model_tree
(
    const SVM_model_tree* tree,
    const char*              model_dirname
);


/** @brief Writes the training data for an SVM model tree to a file. */
Error* write_svm_model_tree_training_data
(
    const Vector_u32* labels, 
    const Matrix_i32* markers,
    const char*       tree_xml_fname,
    const char*       tree_dtd_fname,
    const char*       data_dirname
);


/** @brief Frees a SVM model tree. */
void free_svm_model_tree(SVM_model_tree* tree);


#endif
