/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Weka J48 and PART classifiers.
 */


#ifndef WEKA_H
#define WEKA_H


#include <config.h>

#include <stdlib.h>
#include <inttypes.h>

#include <jwsc/base/error.h>
#include <jwsc/vector/vector.h>
#include <jwsc/matrix/matrix.h>


/** @brief Trains a Weka J48 model and writes it to a file. */
Error* train_weka_j48_model
(
    const Vector_u32* labels, 
    const Matrix_i32* markers,
    const char*       labels_fname,
    const char*       model_fname,
    const char*       weka_jar_fname
);


/** @brief Trains a Weka PART model and writes it to a file. */
Error* train_weka_part_model
(
    const Vector_u32* labels, 
    const Matrix_i32* markers,
    const char*       labels_fname,
    const char*       model_fname,
    const char*       weka_jar_fname
);


/** 
 * @brief Predicts the labels for a set of marker samples using Weka's
 * J48 classifier.
 */
Error* predict_labels_with_weka_j48_model
(
    Vector_u32**      labels_out,
    Vector_d**        confs_out,
    const Matrix_i32* markers,
    const char*       labels_fname,
    const char*       model_fname,
    const char*       weka_jar_fname
);


/** 
 * @brief Predicts the labels for a set of marker samples using Weka's
 * PART classifier.
 */
Error* predict_labels_with_weka_part_model
(
    Vector_u32**      labels_out,
    Vector_d**        confs_out,
    const Matrix_i32* markers,
    const char*       labels_fname,
    const char*       model_fname,
    const char*       weka_jar_fname
);


/** @brief Weka model tree node. */
typedef struct Weka_model_node
{
    /** 
     * @brief Parent of a model node.
     *
     * NULL if the node is the root of the model tree.
     */
    const struct Weka_model_node* parent;

    /** 
     * @brief Parent label for a model node.
     *
     * If the parent is NULL, the value is undefined.
     */
    uint32_t parent_label;

    /** @brief Number of model groups in a node. */
    uint32_t num_groups;

    /** 
     * @brief Model subtrees for each group in a model node.
     * 
     * The ith element is NULL if the ith group does not have a subtree.
     */
    struct Weka_model_node** subtrees;

    /** @brief Labels for each group in a model node. */
    Vector_u32* labels;

    /** @brief Alternative label sets for each group in a model node. */
    Vector_u32** altlabels;

    /** @brief File name for the list of model labels. */
    char* labels_fname;

    /** @brief File name for storage of the model at a node. */
    char* model_fname;
} 
Weka_model_node;


/** @brief Weka model tree. */
typedef  Weka_model_node  Weka_model_tree;


/** @brief Trains a Weka J48 model tree . */
Error* train_weka_j48_model_tree
(
    Weka_model_tree** tree_out,
    const Vector_u32* labels, 
    const Matrix_i32* markers,
    const char*       tree_xml_fname,
    const char*       tree_dtd_fname,
    const char*       model_dirname,
    const char*       weka_jar_fname
);


/** @brief Trains a Weka PART model tree . */
Error* train_weka_part_model_tree
(
    Weka_model_tree** tree_out,
    const Vector_u32* labels, 
    const Matrix_i32* markers,
    const char*       tree_xml_fname,
    const char*       tree_dtd_fname,
    const char*       model_dirname,
    const char*       weka_jar_fname
);


/** 
 * @brief Predicts the labels for a set of marker samples using a Weka J48 model
 * tree.
 */
Error* predict_labels_with_weka_j48_model_tree
(
    Vector_u32**           labels_out,
    Vector_d**             confidence_out,
    const Matrix_i32*      markers,
    const Weka_model_tree* tree,
    const char*            weka_jar_fname
);


/** 
 * @brief Predicts the labels for a set of marker samples using a Weka PART
 * model tree.
 */
Error* predict_labels_with_weka_part_model_tree
(
    Vector_u32**           labels_out,
    Vector_d**             confidence_out,
    const Matrix_i32*      markers,
    const Weka_model_tree* tree,
    const char*            weka_jar_fname
);


/** @brief Reads a Weka model tree. */
Error* read_weka_model_tree
(
    Weka_model_tree** tree_out,
    const char*       tree_xml_fname,
    const char*       tree_dtd_fname,
    const char*       model_dirname
);


/** @brief Frees a Weka model tree. */
void free_weka_model_tree(Weka_model_tree* tree);


#endif
